// CUDA kernel - Double Precision - Real valued
//
//  CUDA kernel for calculating:  local_u = local_u + local_theta*linc + linc[linc_k]
//
//  local_theta : 1 x linc_length = single particle
//  linc        : linc_length x 1 = linear combination
//  linc_k      : array index in linc containing a scalar offset (ignored if set to 0)
//  

__global__ void sabl_collectpart(   const int       batchsize
                                ,   const int       linc_length
                                ,   const double    * theta
                                ,   const double    * linc
                                ,   const int       linc_k
                                ,   double          * u
                                )
{
    //
    // Localise the global-index dependent arrays
    //

    const double  *  local_theta = &theta[threadIdx.x + blockIdx.x*blockDim.x];
    double        *  local_u = u + (threadIdx.x + blockIdx.x*blockDim.x);

    double  accum=0;

    if ( (threadIdx.x + blockIdx.x*blockDim.x) < batchsize)
    {
        const int   warp_limit = 8;     // This must be 2^N, where N is +ve integer
        int     slow_seg;
#if defined (__CUDA_ARCH__)
    #if (__CUDA_ARCH__ >= 300)
        if (batchsize%warp_limit)
        {
            slow_seg = linc_length;
        }
        else
        {
            slow_seg = linc_length%warp_limit;
        }
    #else
        slow_seg = linc_length;
    #endif
#else
        slow_seg = linc_length;
#endif

        int     fast_seg = linc_length - slow_seg;
        int     linc_iter;
        double  part_s[warp_limit];

#if defined (__CUDA_ARCH__)
    #if (__CUDA_ARCH__ >= 300)

// (SM3.x or higher required) use warp shuffling

        int     low32, high32, next_low32, next_high32, ii, the_iter;
        double  elem, shf;

        for ( int kk=0; kk<fast_seg; kk+=warp_limit )
        {
            linc_iter = (kk + (threadIdx.x + blockIdx.x*blockDim.x))%fast_seg;
            const double    * local_linc = linc + linc_iter;

#pragma unroll
            for ( ii=0; ii<warp_limit; ++ii )
            {
                part_s[ii] = 0;
            }

            elem = *local_linc; 
            // part_s[0] +=  elem;

            //
            // Shift from other lanes
            //

#pragma unroll
            for ( ii = 0; ii <=(warp_limit-1); ++ii )
            {
                asm volatile ("mov.b64 {%0,%1}, %2;":"=r"(low32),"=r"(high32):"d"(elem));
                next_low32 = __shfl_xor(low32, ii, warp_limit);
                next_high32 = __shfl_xor(high32, ii, warp_limit);
                asm volatile("mov.b64 %0,{%1,%2};":"=d"(shf):"r"(next_low32),"r"(next_high32));
                part_s[ii ^ (threadIdx.x & (warp_limit-1))] += shf;
            }

#pragma unroll
            for ( ii = 0; ii<warp_limit; ++ii )
            {
                the_iter = linc_iter - (linc_iter & (warp_limit-1)) + ii;
                accum += *(local_theta + batchsize*the_iter) * part_s[ii];
            }
        }
    #endif
#endif

        for ( int i=0; i<slow_seg; ++i)
        {
            part_s[0] = 0;
            linc_iter = fast_seg + (i + (threadIdx.x + blockIdx.x*blockDim.x))%slow_seg;
            const double    * local_linc = linc + linc_iter;
            
            part_s[0] += *local_linc;
            accum += *(local_theta + batchsize*linc_iter) * part_s[0];
        }
        if (linc_k)
        {
            accum += *(linc + linc_k); 
        }
        *(local_u) += accum;
    }
}

