/*
 * This is the CUDA (MEX) code to evaluate likelihood of egarch model. The function operates on model parameters. 
 * 
 * Can survive memory fragmentation (at least for egarch with 5100 observations)
 *
 * To compile:
 *     mex -largeArrayDims -L/usr/local/cuda-6.5/lib64 -lcublas -use_fast_math l_egarch_CUDA.cu
 *     mex -largeArrayDims -L/usr/local/cuda-6.5/lib64 -lcublas l_egarch_CUDA.cu
 *     HX to revisit to add -use_fast_math 4 Sep, 2015  
 *
 * Inputs:
 *   muY_d	mean of y_t (JN x 1), stored linearly 
 *   sigY_d	standard deviation of y_t (JN x 1), stored linearly 
 *   alpha_d	alpha parameters of volatility factors (JN x K), stored linearly in column-major order
 *   beta_d	beta parameters of volatility factors (JN x K), stored linearly in column-major order
 *   gamma_d	gamma parameters of volatility factors (JN x K), stored linearly in column-major order
 *   p_mix_d	parameters p_i of epsilon normal mixture (JN x I), stored linearly in column-major order
 *   mu_mix_d	parameters mu_i of epsilon normal mixture (JN x I), stored linearly in column-major order
 *   sig_mix_d	parameters sig_i of epsilon normal mixture (JN x I), stored linearly in column-major order
 *   y		observed time series y_1:t2 (t2 x 1), stored linearly
 *   intensity	intensity vector
 *   vin_d	v_t1-1 (JN x K), stored linearly in column-major order in global memory
 *   epsin_d	epsilon_t1-1 (JN x 1), stored linearly
 *   JN		the number of particles
 *   K		the number of volatility dynamics
 *   I		the number of epsilon normal mixture
 *   n		length of current information segment	
 *
 * Outputs:
 *   logpy_d	log p(y_t1:t2 | y_1:t1-1, parameters) (JN x 1), stored linearly
 *   vout_d	v_t2 (JN x K), stored linearly in column-major order
 *   epsout_d	epsilon_t2 (JN x 1), stored linearly 
 *
 * Notes:
 *   The argument list needs to be consistent between m_egarch.cu and m_message.m. For more details, see m_message.m and documentation on Matlab function feval().
 *   
 *   The inputs and outputs can be stored either in the main memory (associated with the CPU) or in the global memory (associated with the GPU). To achieve higher efficiency, it is recommended to store them in the global memory.
 *
 */

#ifdef SABL_WIN
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef SABL_LINUX
#include "sys/time.h"
#endif

#include "mex.h"
#include "matrix.h"
#include "gpu/mxGPUArray.h"


#define SQRT_TWO_OVER_PI    0.797884560802865       // sqrt(2/pi)
#define SQRT2PI             2.506628274631000  // sqrt(2*pi)
#define SQRT2               1.414213562373095   // sqrt(2)
#define LOG_SQRT2PI         0.918938533204673   // log( sqrt(2*pi) )
#define KMAX 10
#define IMAX 10

#ifdef SABL_LINUX
double get_wall_time(){
    struct timeval time;
    if (gettimeofday(&time,NULL)){
        //  Handle error
        return 0;
    }
    return (double)time.tv_sec + (double)time.tv_usec * .000001;
}
#endif


template <typename T>
__device__
T normal_cdf( T z ) {
//    return 0.5f * ( 1.0f + erf( z/SQRT2 ));
    return normcdf( z );   // Now a cuda intrinsic
}

template <typename T>
__device__
T normal_cdf( T x, T mu, T sd ) {
    T z = (x-mu)/sd;
//    return 0.5f * ( 1.0f + erf( z/SQRT2 ));
    return normcdf( z );   // Now a cuda intrinsic
}

template <typename T>
__device__ 
T normal_pdf( T z ) {
    return exp( -0.5f*z*z ) / SQRT2PI;
}


template <typename T>
__device__ 
T normal_pdf( T x, T mu, T sd )
{
    T z = (x-mu)/sd;
    return exp( -0.5f*z*z ) / ( sd * SQRT2PI );
}


template <typename T>
__device__ 
T log_normal_pdf( T z )
{
    return -0.5f*z*z  - LOG_SQRT2PI;
}


template <typename T>
__device__ 
T log_normal_pdf( T x, T mu, T sd )
{
    T z = (x-mu)/sd;
    return -0.5f*z*z  - log( sd * SQRT2PI );
}



__global__ 
void evaluate_egarch(	      double * logpy_d, 
			      double * vout_d,
			      double * epsout_d,
                        const double * muY_d, 
                        const double * sigY_d, 
                        const double * alpha_d, 
                        const double * beta_d, 
                        const double * gamma_d, 
                        const double * p_mix_d, 
                        const double * mu_mix_d, 
                        const double * sig_mix_d, 
                        const double * y, 
                        const double * intensity,
			const double * vin_d,
			const double * epsin_d,
				int    JN,
				int    K, 
				int    I, 
				int    n
                    )
{
    int tid = blockIdx.x*blockDim.x + threadIdx.x;
    int nthreads = gridDim.x*blockDim.x;

    // assert K<=KMAX && I<=IMAX

    while (tid < JN)
    {

        double logpy, eps, muY, sigY, 
              v[KMAX], alpha[KMAX], beta[KMAX], gamma[KMAX], 
              p_mix[IMAX], mu_mix[IMAX], sig_mix[IMAX];


        // Get local copies of parameters and states
        eps  =  epsin_d[tid];
        muY  =  muY_d[tid];
       sigY  =  sigY_d[tid];
        for (int k=0; k<K; k++) {
                v[k] =   vin_d[k*JN+tid];
            alpha[k] = alpha_d[k*JN+tid];
             beta[k] =  beta_d[k*JN+tid];
            gamma[k] = gamma_d[k*JN+tid];
        }
        for (int k=0; k<I; k++) {
              p_mix[k] =  p_mix_d[k*JN+tid];
             mu_mix[k] =  mu_mix_d[k*JN+tid];
            sig_mix[k] =  sig_mix_d[k*JN+tid];
        }


        // Evaluate likelihood
        logpy = 0.0;


        for (int t=0; t<n; t++) {
	    double inten = intensity[t];
            if (inten<=1e-6) continue; // 1e-6 instead of 0.0f to allow machine precision error.
            double vsum = 0.0;
            for (int k=0; k<K; k++) {
                v[k] = alpha[k] * v[k] 
                          + beta[k] * ( fabs(eps) - SQRT_TWO_OVER_PI )
                          + gamma[k] * eps;
                vsum += v[k];
            }
            double logh = __logf(sigY) +  0.5 * vsum ;
	    double ysubmuy = y[t] - muY;
	    if (ysubmuy > 0)
	    {
		eps = __expf( __logf( ysubmuy ) - logh );
	    }
	    else
	    {
		eps = -1.0*( __expf( __logf(fabs(ysubmuy)) - logh ) );
	    }
            double pyt = 0.0;
            for (int k=0; k<I; k++) {
                pyt += p_mix[k] * normal_pdf( eps, mu_mix[k], sig_mix[k] );
            }
            logpy += ( __logf( pyt ) - logh ) * inten;
	    
        }
        if ( isnan(logpy) ) logpy = -1.0/0.0; 

        // Save results
        logpy_d[tid] = logpy;
       epsout_d[tid] = eps; 
        for (int k=0; k<K; k++) {
	    vout_d[k*JN+tid] = v[k];
        }

	tid += nthreads;
    }
}



void mexFunction(
		int nlhs, mxArray *plhs[],
		int nrhs, const mxArray *prhs[]
		) 
{
#ifdef _DEBUG_
    double diff, start = get_wall_time();
#endif
    mxInitGPU(); 

    if	(
	    ( 15 != nrhs )
	||  !(mxIsGPUArray(prhs[0]))
	||  !(mxIsGPUArray(prhs[1]))
	||  !(mxIsGPUArray(prhs[2]))
	||  !(mxIsGPUArray(prhs[3]))
	||  !(mxIsGPUArray(prhs[4]))
	||  !(mxIsGPUArray(prhs[5]))
	||  !(mxIsGPUArray(prhs[6]))
	||  !(mxIsGPUArray(prhs[7]))
	||  !(mxIsGPUArray(prhs[8]))
	||  !(mxIsGPUArray(prhs[9]))
	||  !(mxIsGPUArray(prhs[10]))
	||  !(mxIsGPUArray(prhs[11]))
	||  !(mxIsGPUArray(prhs[12]))
	||  !(mxIsGPUArray(prhs[13]))
	||  !(mxIsGPUArray(prhs[14]))
	)
    { 
    	mexErrMsgIdAndTxt("CUDA:MEX:l_egarch_CUDA", "Usage [logp, vout, epsout] = l_egarch_CUDA(muY, sigY, alpha, bet, gam, p, mu, sig, data, intensity, vin, epsin, lopg_rhs, vout_rhs, epsout_rhs). All inputs need to be gpuArray");
    } 

    
    const mxGPUArray * muY_mxGPUArray = mxGPUCreateFromMxArray(prhs[0]);
    const mxGPUArray * sigY_mxGPUArray = mxGPUCreateFromMxArray(prhs[1]);
    const mxGPUArray * alpha_mxGPUArray = mxGPUCreateFromMxArray(prhs[2]);
    const mxGPUArray * bet_mxGPUArray = mxGPUCreateFromMxArray(prhs[3]);
    const mxGPUArray * gam_mxGPUArray = mxGPUCreateFromMxArray(prhs[4]);
    const mxGPUArray * p_mxGPUArray = mxGPUCreateFromMxArray(prhs[5]);
    const mxGPUArray * mu_mxGPUArray = mxGPUCreateFromMxArray(prhs[6]);
    const mxGPUArray * sig_mxGPUArray = mxGPUCreateFromMxArray(prhs[7]);
    const mxGPUArray * data_mxGPUArray = mxGPUCreateFromMxArray(prhs[8]);
    const mxGPUArray * intensity_mxGPUArray = mxGPUCreateFromMxArray(prhs[9]);
    const mxGPUArray * vin_mxGPUArray = mxGPUCreateFromMxArray(prhs[10]);
    const mxGPUArray * epsin_mxGPUArray = mxGPUCreateFromMxArray(prhs[11]);
    mxGPUArray * logp_mxGPUArray = mxGPUCopyFromMxArray(prhs[12]);
    mxGPUArray * vout_mxGPUArray = mxGPUCopyFromMxArray(prhs[13]);
    mxGPUArray * epsout_mxGPUArray = mxGPUCopyFromMxArray(prhs[14]);

    if (
	    mxGPUGetClassID(muY_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(sigY_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(alpha_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(bet_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(gam_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(p_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(mu_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(sig_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(data_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(intensity_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(vin_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(epsin_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(logp_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(vout_mxGPUArray) != mxDOUBLE_CLASS
	 || mxGPUGetClassID(epsout_mxGPUArray) != mxDOUBLE_CLASS
	)
    {
	mexErrMsgIdAndTxt("CUDA:MEX:l_egarch_CUDA", "Usage [logp, vout, epsout] = l_egarch_CUDA(muY, sigY, alpha, bet, gam, p, mu, sig, data, intensity, vin, epsin, logp_rhs, vout_rhs, epsout_rhs). All inputs need to be gpuArray with the underlying class of DOUBLE");
    }

    const mwSize * dims_alpha = mxGPUGetDimensions(alpha_mxGPUArray);
    const mwSize * dims_p = mxGPUGetDimensions(p_mxGPUArray);
    int JNwork = dims_alpha[0];
    int K = dims_alpha[1];
    int I = dims_p[1];
    int n = mxGPUGetNumberOfElements(intensity_mxGPUArray);
	
    const double * muY = (double const *)mxGPUGetDataReadOnly(muY_mxGPUArray);	
    const double * sigY = (double const *)mxGPUGetDataReadOnly(sigY_mxGPUArray);	
    const double * alpha = (double const *)mxGPUGetDataReadOnly(alpha_mxGPUArray);	
    const double * bet = (double const *)mxGPUGetDataReadOnly(bet_mxGPUArray);	
    const double * gam = (double const *)mxGPUGetDataReadOnly(gam_mxGPUArray);	
    const double * p = (double const *)mxGPUGetDataReadOnly(p_mxGPUArray);	
    const double * mu = (double const *)mxGPUGetDataReadOnly(mu_mxGPUArray);	
    const double * sig = (double const *)mxGPUGetDataReadOnly(sig_mxGPUArray);	
    const double * data = (double const *)mxGPUGetDataReadOnly(data_mxGPUArray);	
    const double * intensity = (double const *)mxGPUGetDataReadOnly(intensity_mxGPUArray);	
    const double * vin = (double const *)mxGPUGetDataReadOnly(vin_mxGPUArray);	
    const double * epsin = (double const *)mxGPUGetDataReadOnly(epsin_mxGPUArray);	
	  double * logp = (double *)mxGPUGetData(logp_mxGPUArray);
	  double * vout = (double *)mxGPUGetData(vout_mxGPUArray);	
	  double * epsout = (double *)mxGPUGetData(epsout_mxGPUArray);	

    evaluate_egarch <<<128, 128>>> ( 	logp, 
			vout,
			epsout,
                        muY, 
                        sigY, 
                        alpha, 
                        bet, 
                        gam, 
                        p, 
                        mu, 
                        sig, 
                        data, 
                        intensity,
			vin,
			epsin,
                        JNwork,
                        K, 
                        I, 
                        n
                    );

    plhs[0] = mxGPUCreateMxArrayOnGPU(logp_mxGPUArray); 
    if (logp_mxGPUArray) {mxGPUDestroyGPUArray(logp_mxGPUArray); logp_mxGPUArray = NULL;}

    plhs[1] = mxGPUCreateMxArrayOnGPU(vout_mxGPUArray); 
    if (vout_mxGPUArray) {mxGPUDestroyGPUArray(vout_mxGPUArray); vout_mxGPUArray = NULL;}

    plhs[2] = mxGPUCreateMxArrayOnGPU(epsout_mxGPUArray);
    if (epsout_mxGPUArray) {mxGPUDestroyGPUArray(epsout_mxGPUArray); epsout_mxGPUArray = NULL;}

    if (muY_mxGPUArray) {mxGPUDestroyGPUArray(muY_mxGPUArray); muY_mxGPUArray = NULL;}
    if (sigY_mxGPUArray) {mxGPUDestroyGPUArray(sigY_mxGPUArray); sigY_mxGPUArray = NULL;}	
    if (alpha_mxGPUArray) {mxGPUDestroyGPUArray(alpha_mxGPUArray); alpha_mxGPUArray = NULL;}	
    if (bet_mxGPUArray) {mxGPUDestroyGPUArray(bet_mxGPUArray); bet_mxGPUArray = NULL;}	
    if (gam_mxGPUArray) {mxGPUDestroyGPUArray(gam_mxGPUArray); gam_mxGPUArray = NULL;}
    if (p_mxGPUArray) {mxGPUDestroyGPUArray(p_mxGPUArray); p_mxGPUArray = NULL;}
    if (mu_mxGPUArray) {mxGPUDestroyGPUArray(mu_mxGPUArray); mu_mxGPUArray = NULL;}	
    if (sig_mxGPUArray) {mxGPUDestroyGPUArray(sig_mxGPUArray); sig_mxGPUArray = NULL;}
    if (data_mxGPUArray) {mxGPUDestroyGPUArray(data_mxGPUArray); data_mxGPUArray = NULL;}	
    if (intensity_mxGPUArray) {mxGPUDestroyGPUArray(intensity_mxGPUArray); intensity_mxGPUArray = NULL;}	
    if (vin_mxGPUArray) {mxGPUDestroyGPUArray(vin_mxGPUArray); vin_mxGPUArray = NULL;}
    if (epsin_mxGPUArray) {mxGPUDestroyGPUArray(epsin_mxGPUArray); epsin_mxGPUArray = NULL;}

}
