// CUDA kernel - Double Precision - Real valued
//
// Calculate the sign and magnitude of the determinant of matrix A that
// has been factorised A = P'LU. 
//
// The sum of the log of diagonal elements in
// u (which is U from P'LU factorisation in the upper triangular portion)
// equals the magnitude of the determinant.
//
// The first element of u (relevant to a given thread/particle) is over-
// written with the magnitude of log(determinant)
//
// The second element of u (relevant to a given thread/particle) is over-
// written with the sign of the determinant
//

__global__ void sabl_logdet(    const int       batchsize,
                                const int       myN,
                                const int     * p,
                                double        * u
                            )
{
    double  ln_magnitude=0;
    int     sign=1;
    int     p_i, 
            p_j;

    int          thread_memoffset = (threadIdx.x + blockIdx.x*blockDim.x)*myN*myN;
    double *     local_u = u + thread_memoffset;
    const int *  local_p = p + (threadIdx.x + blockIdx.x*blockDim.x)*myN;

    if ( (threadIdx.x + blockIdx.x*blockDim.x) < batchsize)
    {
        for (p_i=0; p_i<myN; ++p_i)
        {
            ln_magnitude += log( fabs(*(local_u)) );
            sign = ( signbit(*(local_u)) != 0 )  ? -sign : sign;
            local_u = local_u + myN + 1;
        }

        // Count the number of row swaps in the vector p

        for (p_j=0; p_j<myN; ++p_j)
        {
            sign = ( local_p[p_j] != p_j+1 ) ? -sign : sign;
        }

        *(u + thread_memoffset) = ln_magnitude;
        *(u + thread_memoffset + 1) = sign;
    }
}

