/*
 * C++, CUDA and MEX headers come first
 */
#define WIN32_LEAN_AND_MEAN
#include <math.h>
#include <mex.h>
#include <gpu/mxGPUArray.h>
#include <cublas_v2.h>
#include <stdint.h>

/*
 * General SABL headers go in the next section
 */
#ifndef MVN_H
#include <l_mvn.h>
#endif

/*
 * C/CUDA code from the SABL Library gets included here
 */
#include <l_logdet.cu>
#include <l_collectpart.cu>

/*
 * C/CUDA code for a particular SABL Model gets included here
 */
#include <m_thetapar.cu>


/*
 * C/CUDA code that is SABL Project dependent gets included here
 */

/*
#include <p_thetapar.cu>
 */

/*
 * CUDA kernels
 */

//
// MVN with sufficient statistics.
// d_data holds the F matrix:  F'*F = [Y X]' * Intensity * [Y X]
// Intensity is handled during the calculation of F
//
__global__ void sabl_mvn  ( //
                            // Particle-related arguments
                            //
                            const double *    theta
                          , const int         batchsize
                          , const int         startrow_H
                          , const int         endrow_H
                            //
                            // Extended theta for non-linear transforms
                            //
                          , const double *    theta_extended
                            //
                            // Data-related arguments
                            //
                          , const double *    d_data
                          , const int         data_observations 
                          , const int         data_vectorsize
                            //
                            //  Return results
                            //
                          , double *          result
                          )
/*  
 *   Kernel for calculating: local_result = (d_data * [(I-A)'; -B'] * H').^2
 *
 *   Assumes the following globally declared and initialised constant memory locations
 *
 *   #define X_K;
 *   #define Y_N;
 *   #define THREADS_PER_BLOCK;
 */

{
    //
    // Localise the local (block) index dependent arrays
    //

    double 		acc[X_K + Y_N];
    double 	*	mybase_acc = &acc[0];


    //
    // Localise the global-index dependent arrays
    //
    const double  *  local_theta = &theta[threadIdx.x + blockIdx.x*blockDim.x];
    const double  *  local_theta_extended = &theta_extended[threadIdx.x + blockIdx.x*blockDim.x];
    double        *  local_result = result + (threadIdx.x + blockIdx.x*blockDim.x);



    // Form [(I-A)'; -B']*H'

    uint                iter_list2_H,
                        row_value_H,
                        iter_list2,
                        row_index_AB,
                        column_index_AB;

    double              e_H = 0;
    double              partial_pdf = 0;

    const int *        ptr_list2_H = &list2H_rt[0];

   if ( (threadIdx.x + blockIdx.x*blockDim.x) < batchsize)
   {
        for ( row_value_H=startrow_H; row_value_H <= endrow_H; ++row_value_H )
        {
            // Initialise the shared memory array: mybase_acc
            for (uint i=0; i<(X_K_rt + Y_N_rt); ++i)
            {
                *(mybase_acc + i)=0;
            }

            for ( iter_list2_H=0; iter_list2_H < 2*Hsize_rt; iter_list2_H+=2 )
            {
                const uint row_index_H      = *(ptr_list2_H + iter_list2_H);     // 1-based row index of H
                const uint column_index_H   = *(ptr_list2_H + iter_list2_H + 1); // 1-based column index of H

                if ( row_index_H == row_value_H )
                {
                    e_H = m_thetapar_H(iter_list2_H>>1
					, row_index_H
					, column_index_H
					, local_theta
					, local_theta_extended
					, batchsize);   

                    // Start with matrix -B'*H'

                    const int * ptr_list2_AB = &list2B_rt[0];

                    for ( iter_list2=0; iter_list2 < 2*Bsize_rt; iter_list2+=2 )
                    {
                        row_index_AB = *(ptr_list2_AB + iter_list2);
                        if ( row_index_AB == column_index_H )
                        {
                            if ( column_index_H <= Y_N_rt )
                            {
                            	column_index_AB = *(ptr_list2_AB + iter_list2 + 1);

                            	// We have reached a non-zero product...so accumulate it!

                            	*(mybase_acc + (Y_N_rt + column_index_AB-1)) += ( -m_thetapar_B(    iter_list2>>1
                                                                                                , row_index_AB
                                                                                                , column_index_AB
                                                                                                , local_theta
                                                                                                , local_theta_extended
                                                                                                , batchsize
                                                                                                )
                                                                                ) * e_H; 

                            }
                        }
                    }
                    // Now do matrix (I-A)'*H'
                    if ( column_index_H <= Y_N_rt )
                    {

                        *(mybase_acc + (column_index_H-1)) += e_H;

                    }

                    ptr_list2_AB = &list2A_rt[0];
                    for ( iter_list2=0; iter_list2 < 2*Asize_rt; iter_list2+=2 )
                    {
                        row_index_AB = *(ptr_list2_AB + iter_list2);
                        if ( row_index_AB == column_index_H )
                        {
                            if ( column_index_H <= Y_N_rt )
                            {
                            	column_index_AB = *(ptr_list2_AB + iter_list2 + 1);

                                // We have reached a non-zero product...so accumulate it!

                            	*(mybase_acc + (column_index_AB-1)) += ( - m_thetapar_A(  iter_list2>>1
                                                                                        , row_index_AB
                                                                                        , column_index_AB
                                                                                        , local_theta
                                                                                        , local_theta_extended
                                                                                        , batchsize
                                                                                        )
                                                                       ) * e_H;

                            } 
                        }
                    }
                }
            }

            const int   warp_limit = 8;
            int         slow_seg;

#if defined (__CUDA_ARCH__)
    #if (__CUDA_ARCH__ >= 300)
        if (batchsize%warp_limit)
        {
                slow_seg = data_observations;
        }
        else
        {
                 slow_seg = data_observations%warp_limit;
        }
    #else
                 slow_seg = data_observations;
    #endif
#else
                 slow_seg = data_observations;
#endif
            int     fast_seg = data_observations - slow_seg;
            double  part_s[warp_limit];
        
#if defined (__CUDA_ARCH__)
    #if (__CUDA_ARCH__ >= 300)

// (SM3.x or higher required) use warp shuffling
   
            int     low32, high32, next_low32, next_high32, ii;
            double  shf;
  
            for ( int kk=0; kk<fast_seg; kk+=warp_limit )
            {
                const double    * local_data = d_data + ( (kk + (threadIdx.x + blockIdx.x*blockDim.x))%fast_seg );
#pragma unroll
                for ( ii=0; ii<warp_limit; ++ii )
                {
                    part_s[ii] = 0;
                }
#pragma unroll
                for ( iter_list2=0; iter_list2 < (X_K_rt + Y_N_rt); ++iter_list2 )
                {
                    e_H = *(local_data + data_observations*iter_list2); 
                    part_s[0] +=  e_H *
                                    *(mybase_acc + iter_list2);
                    //
                    // Shift from other lanes
                    //
#pragma unroll
                    for ( ii = 1; ii <=(warp_limit-1); ++ii )
                    {
                        asm volatile ("mov.b64 {%0,%1}, %2;":"=r"(low32),"=r"(high32):"d"(e_H));
                        next_low32 = __shfl_xor(low32, ii, warp_limit);
                        next_high32 = __shfl_xor(high32, ii, warp_limit);
                        asm volatile("mov.b64 %0,{%1,%2};":"=d"(shf):"r"(next_low32),"r"(next_high32));

                        part_s[ii] += shf *     
                                        *(mybase_acc + iter_list2);
                    }
                }

                partial_pdf += part_s[0] * part_s[0];

#pragma unroll
                for ( ii = 1; ii <=(warp_limit-1); ++ii )
                {
                    partial_pdf += part_s[ii] * part_s[ii];
                }
            }
    #endif
#endif

            for ( int i=0; i<slow_seg; ++i)
            {
                    part_s[0] = 0;
                    const double    * local_data = d_data + fast_seg + ( (i+threadIdx.x + blockIdx.x*blockDim.x)%slow_seg );

#pragma unroll
                    for ( iter_list2=0; iter_list2 < (X_K_rt + Y_N_rt); ++iter_list2 )
                    {
                        part_s[0] +=  *(local_data + data_observations*iter_list2) *
                                        *(mybase_acc + iter_list2);
                    }

                    partial_pdf += part_s[0] * part_s[0];
            }
        }
    *local_result = partial_pdf;
    }
} 

__global__ void sabl_mvn_getI_minusA( const double * theta
                                    , const double * theta_extended
                                    , const int batchsize
                                    , const int myN
                                    , double * result 
                                    )
{
    //
    // Localise the global-index dependent arrays
    //
    const double  *  local_theta = &theta[threadIdx.x + blockIdx.x*blockDim.x];
    const double  *  local_theta_extended = &theta_extended[threadIdx.x + blockIdx.x*blockDim.x];
    double        *  local_result = result + (threadIdx.x + blockIdx.x*blockDim.x)*myN*myN;


    if ( (threadIdx.x + blockIdx.x*blockDim.x) < batchsize)
    {
        uint	iter_list2_A,
                row_index_A,
                column_index_A;

        const int * ptr_list2_A = &list2A_rt[0];
        for ( iter_list2_A=0; iter_list2_A < 2*Asize_rt; iter_list2_A+=2 )
        {
                        row_index_A = *(ptr_list2_A + iter_list2_A);
                        column_index_A = *(ptr_list2_A + iter_list2_A + 1);
                        if ( (row_index_A <= myN) && (column_index_A <= myN) )
                        {
                            *(local_result + myN*(column_index_A - 1) + (row_index_A - 1)) = -m_thetapar_A(   iter_list2_A>>1
                                                                                                            , row_index_A
                                                                                                            , column_index_A
                                                                                                            , local_theta
                                                                                                            , local_theta_extended
                                                                                                            , batchsize
                                                                                                            );

                          // For debugging: prints the row (*100) + column indices  
                          // *(local_result + myN*(column_index_A-1) + (row_index_A-1)) = row_index_A*100 + column_index_A;
                        }
        }
        #pragma unroll
        for ( iter_list2_A=0; iter_list2_A < myN; ++iter_list2_A )
        {
            *(local_result + myN*iter_list2_A + iter_list2_A) += 1;
        }
    }	
}

__global__ void sabl_mvn_get_logDetH( const double * theta
                                    , const double * theta_extended
                                    , double * result
                                    , const int myN
                                    , const int batchsize
                                    )
{
    //
    // Localise the global-index dependent arrays
    //
    const double  *  local_theta = &theta[threadIdx.x + blockIdx.x*blockDim.x];
    const double  *  local_theta_extended = &theta_extended[threadIdx.x + blockIdx.x*blockDim.x];
    double        *  local_result = result + (threadIdx.x + blockIdx.x*blockDim.x);

    if ( (threadIdx.x + blockIdx.x*blockDim.x) < batchsize)
    {
        uint	iter_list2_H,
                row_index_H,
                column_index_H;

        double  accum=0;
        const int * ptr_list2_H = &list2H_rt[0];
        for ( iter_list2_H=0; iter_list2_H < 2*Hsize_rt; iter_list2_H+=2 )
        {
                        row_index_H = *(ptr_list2_H + iter_list2_H);
                        column_index_H = *(ptr_list2_H + iter_list2_H + 1);
                        if (row_index_H == column_index_H)
                        {
                            if (column_index_H <= myN)
                            {
                                accum += log( m_thetapar_H(   iter_list2_H>>1
                                                            , row_index_H
                                                            , column_index_H
                                                            , local_theta
                                                            , local_theta_extended
                                                            , batchsize) );
                            }
                        }
        }
    *local_result += accum;
    }	
}

__global__ void sabl_mvn_gatherpdf(	  const int batchsize
                                    , const double T
                                    , const int N
                                    , double *result
                                    , const double *partial_logdetA
                                    , const double *partial_V
                                    )
{   
    const double PI=3.141592653589793;

    //
    // Localise the global-index dependent arrays
    //
    double              *   local_result = result + (threadIdx.x + blockIdx.x*blockDim.x);
    const double        *   local_partial_logdetA = partial_logdetA + (threadIdx.x + blockIdx.x*blockDim.x)*N*N;
    const double        *   local_partial_V = partial_V + (threadIdx.x + blockIdx.x*blockDim.x);

    double              accum=0;

    if ( (threadIdx.x + blockIdx.x*blockDim.x) < batchsize)
    {
        accum = T*(-0.5*N*log(2*PI) + *local_result + *local_partial_logdetA);
        for (int lh=0; lh<N; ++lh)
        {
            accum += -0.5 * *(local_partial_V + lh*batchsize);
        }
        *local_result = accum;
    }
}


/*
 * Host code
 */

void cleanup()
{
	mexPrintf("Cleanup completed.\n");
}

void convert_with_linc(     const mxArray           * M
                            ,   int                 * M_l2
                            ,   int                 * M_map
                            ,   int                 * M_mapsize
                            ,   const int           M_base_in_te_v
                            ,   const mxGPUArray * Mlinc
                            ,   const int       Mlinc_has_offset
                            ,   const int       batch
                            ,   const mxGPUArray   * theta
                            ,   double          * d_te_v
                            ,   cudaStream_t    & my_stream
                            )
{
    double const * M_map_ptr = (double const *)(mxGetData(M));
    int blocksPerGrid = (batch + THREADS_PER_BLOCK - 1) / THREADS_PER_BLOCK;

    int nz_l2 = 0;
    int sparsechk = 0;
    int qty_rows = (int)mxGetM(M);
    int offset_in_te_v = 0;

    mwSize const * Mlinc_size = mxGPUGetDimensions(Mlinc);  
    mwSize const * theta_size = mxGPUGetDimensions(theta);
    double const * Mlinc_ptr = (double const *)(mxGPUGetDataReadOnly(Mlinc));
    double const * d_theta = (double const *)(mxGPUGetDataReadOnly(theta));

    for ( int jn = 0; jn < (int)mxGetN(M); ++jn)
    {
        for ( int im=0; im < qty_rows; ++im)
        {
            sparsechk = (int)( *(M_map_ptr + qty_rows*jn + im) );
            if (sparsechk != 0)
            {
                *M_mapsize = nz_l2>>1; 
                *(M_l2 + nz_l2) = im+1;            
                *(M_l2 + nz_l2+1) = jn+1;
                if (sparsechk > 0 )
                {
                    *(M_map + (nz_l2>>1)) = sparsechk;	
                }
                else
                {
                    *(M_map + (nz_l2>>1)) = -( M_base_in_te_v + offset_in_te_v);

                    int effective_lincsize = theta_size[1];
                    int effective_offsetidx = 0;

                    if (Mlinc_has_offset)
                    {
                        effective_offsetidx = theta_size[1];
                    }

                    sabl_collectpart<<<     blocksPerGrid
                                    ,       THREADS_PER_BLOCK     
                                    ,       0
                                    ,       my_stream
                                    >>>(    batch
                                        ,   effective_lincsize
                                        ,   d_theta
                                        ,   Mlinc_ptr + Mlinc_size[0]*(abs(sparsechk) - 1)
                                        ,   effective_offsetidx
                                        ,   d_te_v + batch*(M_base_in_te_v + offset_in_te_v - 1)
                                        );
                    offset_in_te_v +=1;
                }

                // mexPrintf("M: %d,%d : %d\n", *(M_l2 + nz_l2), *(M_l2 + nz_l2+1), *(M_map + (nz_l2>>1)) );

                nz_l2 += 2;
            }
        }
    }
    mxFree((void *)Mlinc_size);
    mxFree((void *)theta_size);

}

void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, mxArray const *prhs[])
{
    /*
     * prhs[0] : theta 	
                    	* GPUArray - matrix with JN rows, 1 row per 
                        * particle.  Positive values in maps represent
                        * (1-based) index of columns in theta
     * prhs[1] : mydata 
                    	* GPUArray - this will hold the F matrix of
                        * covariates (X) and outcomes (Y):
                        *  F'*F = [Y X]'*diag(intensity)*[Y X]

     * prhs[2] : T      
                        * T (sum of the intensity vector)

     * prhs[3] : Amap 		
                        * N x N matrix - +ve values refer to the (1-based)
                        * column index in theta.  
                        * -ve values are converted to their absolute
                        * value, after which they refer to the 
                        * (1-based) column index in Alinc.  

     * prhs[4] : Alinc
                        * GPUArray
                        * A = theta * Alinc(:, Amap(find(Amap(:)<0)) )

     * prhs[5] : Aname
                        * Name string for the matrix described by Amap, Alinc

     * prhs[6] : Bmap
                        * N x K matrix that maps matrix B in the MVN model.
                        * Similar to Amap described above.

     * prhs[7] : Blinc
                		* GPUArray
                        * B = theta * Blinc(:, Bmap(find(Bmap(:)<0)) )

     * prhs[8] : Bname
                		* Name string for the matrix described by Bmap, Blinc
     
     * prhs[9] : Hmap
                        * N x N matrix that maps matrix H in the MVN model.
                        * Similar to Amap described above.
     * prhs[10] : Hlinc
                		* GPUArray
                        * H = theta * Hlinc(:, Hmap(find(Hmap(:)<0)) )

     * prhs[11] : Hname
                		* Name string for the matrix described by Hmap, Hlinc
    
     * prhs[12] : A_tri - Not used
     * prhs[13] : pdf 	
                        * GPUArray - initial values and return values - 
                        * column vector JN x 1 
                        * for log-likelihood evaluations

     * prhs[14] : theta_extended 
                        * GPUArray 
                        * JN rows x #columns needed for all linear and 
                        * non-linear numerical results derived from theta
     */

    char  compiled_k[8];
    sprintf(compiled_k,  "%d", X_K);

    char  compiled_n[8];
    sprintf(compiled_n,  TOSTRING( Y_N ) );

    char  compiled_tpb[32];
    sprintf(compiled_tpb,  TOSTRING( THREADS_PER_BLOCK ) );

    char  compiled_list2y[1024];
    sprintf(compiled_list2y,  TOSTRING((1, LIST2_Y)) );

    char  compiled_ymap[1024];
    sprintf(compiled_ymap,  TOSTRING((Y_MAP)) );

    char  compiled_list2x[1024];
    sprintf(compiled_list2x,  TOSTRING((1, LIST2_X)) );

    char  compiled_xmap[1024];
    sprintf(compiled_xmap,  TOSTRING((X_MAP)) );

    char  compiled_list2A[1024];
    sprintf(compiled_list2A,  TOSTRING((LIST2_A)) );

    char  compiled_Amap[1024];
    sprintf(compiled_Amap,  TOSTRING((A_MAP)) );

    char  compiled_list2B[1024];
    sprintf(compiled_list2B,  TOSTRING((LIST2_B)) );

    char  compiled_Bmap[1024];
    sprintf(compiled_Bmap,  TOSTRING((B_MAP)) );

    char  compiled_list2H[1024];
    sprintf(compiled_list2H,  TOSTRING((LIST2_H)) );

    char  compiled_Hmap[1024];
    sprintf(compiled_Hmap,  TOSTRING((H_MAP)) );

    const char * compiled_all[13];

    compiled_all[0] = (char *)compiled_k;
    compiled_all[1] = (char *)compiled_n;
    compiled_all[2] = (char *)compiled_tpb;
    compiled_all[3] = (char *)compiled_list2x;
    compiled_all[4] = (char *)compiled_xmap;
    compiled_all[5] = (char *)compiled_list2y;
    compiled_all[6] = (char *)compiled_ymap;
    compiled_all[7] = (char *)compiled_list2A;
    compiled_all[8] = (char *)compiled_Amap;
    compiled_all[9] = (char *)compiled_list2B;
    compiled_all[10] = (char *)compiled_Bmap;
    compiled_all[11] = (char *)compiled_list2H;
    compiled_all[12] = (char *)compiled_Hmap;

    /* Declare all variables.*/
    mxGPUArray const    *   theta
                    , 	*   theta_extended
                    ,   *   mydata
                    ,   *   Alinc
                    ,   *   Blinc
                    ,   *   Hlinc
                    ,	*   pdf
                    ;

    mxGPUArray 	*	ret_pdf
            ,   *   theta_extended_volatile
            ;

    int             batch;

    char const * const errId = "parallel:gpu:mexGPUExample:InvalidInput";
    char const * const errMsg = "Invalid input to MEX file. Check that all Maps are stored on the host memory.";
    char const * const errIdCuda = "parallel:gpu:mexGPUExample:InvalidInput";
    char const * const errMsgCuda = "Failure in cudaMemcpyToSymbol";

    cudaStream_t            stream[3];

    cublasStatus_t          stat;
    cublasHandle_t          handle;
    cudaError_t             cu_stat;

    double          ** 	    h_A,
                    ** 	    d_suB;

    int             *       pivots_d,
                    *       opstatus_d;
    if (nrhs>0)
    {

        if ((nrhs!=15) 
        /* Throw an error if the particles or the observations are not a GPU array. */
            || !(mxIsGPUArray(prhs[0])) 
            || !(mxIsGPUArray(prhs[1]))
            || !(mxIsGPUArray(prhs[13])) 
            || !(mxIsGPUArray(prhs[14]))
         /* Note however that maps need to be on CPU memory! */
            || (mxIsGPUArray(prhs[3])) 
            || (mxIsGPUArray(prhs[6])) 
            || (mxIsGPUArray(prhs[9])) 
 
            ) 
        {
            mexErrMsgIdAndTxt(errId, errMsg);
        }


        /* Initialize the MathWorks GPU API. */
        mxInitGPU();
        stat = cublasCreate(&handle);
        if (stat != CUBLAS_STATUS_SUCCESS)
        {
               if (stat == CUBLAS_STATUS_NOT_INITIALIZED )
                {
                    mexErrMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check that cublasCreate was called.");
                    mexAtExit(cleanup);
                }
               if (stat == CUBLAS_STATUS_ALLOC_FAILED )
                {
                    mexErrMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check that GPU resource has been freed.");
                    mexAtExit(cleanup);
                }
        }

        theta 		= mxGPUCreateFromMxArray(prhs[0]);
        mydata 		= mxGPUCreateFromMxArray(prhs[1]);

        Alinc       = mxGPUCreateFromMxArray(prhs[4]);
        Blinc       = mxGPUCreateFromMxArray(prhs[7]);
        Hlinc       = mxGPUCreateFromMxArray(prhs[10]);

        pdf  		= mxGPUCreateFromMxArray(prhs[13]);
        theta_extended 	= mxGPUCreateFromMxArray(prhs[14]);


        mwSize const *matrixdim =   mxGPUGetDimensions(theta); 
        mwSize const * Alinc_size = mxGPUGetDimensions(Alinc);  
        mwSize const * Blinc_size = mxGPUGetDimensions(Blinc);  
        mwSize const * Hlinc_size = mxGPUGetDimensions(Hlinc);  

        int Alinc_has_offset = (Alinc_size[0] > matrixdim[1]) ? 1 : 0;
        int Blinc_has_offset = (Blinc_size[0] > matrixdim[1]) ? 1 : 0;
        int Hlinc_has_offset = (Hlinc_size[0] > matrixdim[1]) ? 1 : 0;

        int A_base_in_te_v = 1;
        int B_base_in_te_v = A_base_in_te_v + Alinc_size[1];
        int H_base_in_te_v = B_base_in_te_v + Blinc_size[1];

        if ( mxGPUGetNumberOfDimensions(theta) < 2 )
        {
            batch=1;
        }
        else
        {
            batch = matrixdim[0];   // the number of rows = number of particles 
        }

        if ( (batch < 1) )
        {
            mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        }
        else
        {
              int blocksPerGrid = (batch + THREADS_PER_BLOCK - 1) / THREADS_PER_BLOCK;

            /*
             * Verify that theta really is a double array before extracting the pointer.
             */
               if (mxGPUGetClassID(theta) != mxDOUBLE_CLASS) {
                   mexErrMsgIdAndTxt(errId, errMsg);
                }

            /*
             * Verify that theta_extended really is a double array before extracting the pointer.
             */
               if (mxGPUGetClassID(theta_extended) != mxDOUBLE_CLASS) {
                   mexErrMsgIdAndTxt(errId, errMsg);
                }
            /*
             * Now that we have verified the data type, extract a pointer to the input
             * data on the device.
             */

               double const * d_theta = (double const *)(mxGPUGetDataReadOnly(theta));
               double const * d_theta_extended = (double const *)(mxGPUGetDataReadOnly(theta_extended));
               double const * d_mydata = (double const *)(mxGPUGetDataReadOnly(mydata));
               double const * T_ptr 	= (double const *)(mxGetData(prhs[2]));

               double const * Amap_ptr 	= (double const *)(mxGetData(prhs[3]));
               double const * Alinc_ptr = (double const *)(mxGPUGetDataReadOnly(Alinc));
               char   const * Aname_ptr = (char const *)(mxGetData(prhs[5]));

               double const * Bmap_ptr 	= (double const *)(mxGetData(prhs[6]));
               double const * Blinc_ptr = (double const *)(mxGPUGetDataReadOnly(Blinc));
               char   const * Bname_ptr = (char const *)(mxGetData(prhs[8]));

               double const * Hmap_ptr 	= (double const *)(mxGetData(prhs[9]));
               double const * Hlinc_ptr = (double const *)(mxGPUGetDataReadOnly(Hlinc));
               char   const * Hname_ptr = (char const *)(mxGetData(prhs[11]));

               int X_K_set_constant[1];
               int Y_N_set_constant[1];

               double const * A_tri_ptr = (double const *)(mxGetData(prhs[12]));
               double const * d_pdf 	= (double const *)(mxGPUGetDataReadOnly(pdf));
    
            /*
             *
             * Scratch space 
             *
             */

            /* 
             *  Create a GPUArray to hold intermediate results 
             *  and get its underlying pointer. 
             */

               matrixdim = mxGPUGetDimensions(theta_extended);
               theta_extended_volatile = mxGPUCreateGPUArray(mxGPUGetNumberOfDimensions(theta_extended),
                                    mxGPUGetDimensions(theta_extended),
                                    mxGPUGetClassID(theta_extended),
                                    mxGPUGetComplexity(theta_extended),
                                    MX_GPU_INITIALIZE_VALUES);
               double  * d_te_v = (double *)(mxGPUGetData(theta_extended_volatile));

            /* 
             * Initialise volatile values to be the same as input values 
             */

               cudaMemcpy( d_te_v,
                           d_theta_extended,
                           sizeof(double) * matrixdim[0] * matrixdim[1],
                           cudaMemcpyDeviceToDevice);

            /*
             * Returned values 
             */

            /* Create a GPUArray to hold the result and get its underlying pointer. */
              
               ret_pdf = mxGPUCreateGPUArray(mxGPUGetNumberOfDimensions(pdf),
                                    mxGPUGetDimensions(pdf),
                                    mxGPUGetClassID(pdf),
                                    mxGPUGetComplexity(pdf),
                                    MX_GPU_INITIALIZE_VALUES);
               double  * d_ret_pdf = (double *)(mxGPUGetData(ret_pdf));

            /* Initialise return values to be the same as input values */


               cudaMemcpy( d_ret_pdf,
                           d_pdf,
                           sizeof(double) * batch,
                           cudaMemcpyDeviceToDevice);

            /* 
             * dbg:  to construct: log(det(I-A))
             * trsum: to calculate: [Y X]*[ (I-A)'; -B' ]*H'
             */

            cudaGetLastError();


            X_K_set_constant[0] = (int)mxGetN(prhs[6]);
            Y_N_set_constant[0] = (int)mxGetN(prhs[3]);

            mwSize  const   dbg_dims[] = {Y_N_set_constant[0], Y_N_set_constant[0], batch};
            // mwSize  const   trsum_dims[] = {batch, 1};
            mwSize  const   trsum_dims[] = {batch, Y_N_set_constant[0]};

            mxGPUArray * dbg = mxGPUCreateGPUArray(      3,
                                                        dbg_dims,
                                                        mxDOUBLE_CLASS,
                                                        mxREAL,
                                                        MX_GPU_INITIALIZE_VALUES);
            double * d_dbg = (double *)( mxGPUGetData(dbg) );

            mxGPUArray * trsum = mxGPUCreateGPUArray(   2,
                                                        trsum_dims,
                                                        mxDOUBLE_CLASS,
                                                        mxREAL,
                                                        MX_GPU_INITIALIZE_VALUES);
            double * d_trsum = (double *)( mxGPUGetData(trsum) );


            cudaMalloc((void**) &pivots_d, sizeof(int) * batch * Y_N_set_constant[0]);
            cudaMalloc((void**) &opstatus_d, sizeof(int)*batch);

            cudaMallocHost((void***) &h_A, sizeof(double *) * batch);
            cudaMalloc((void***) &d_suB, sizeof(double *) * batch);
            cudaDeviceSynchronize();

            /*
             * Calculate: sum( log(h(i,i)) ) along the diagonal of matrix H
             */

            cudaStreamCreate(&stream[0]);
            cudaGetLastError();

            // 
            //  Process matrix: H
            //  Transfer to GPU
            // - the list of non-zero elements for H: list2H_rt
            // - the quantity of such elements: Hsize_rt
            // - the list of mappings for H: list_Hmap_rt
            // 
            int Hmap_rt[Y_N*Y_N];
            int H_sizert[1];	
            int l2_H_rt[2*Y_N*Y_N];

            // cu_stat = cudaMemcpyFromSymbol( l2_H_rt, list2H_rt, 2*Y_N*Y_N*sizeof(int));
            //     if (cu_stat != cudaSuccess)
            //     {
            //         mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
            //     }
            //     else
            //     {
            //         for ( uint in = 0; in < 2*Y_N*Y_N; ++in )
            //         {
            //             mexPrintf("Compiled list2H_rt: %d\n", l2_H_rt[in]);
            //         }
            //     }


            convert_with_linc(      prhs[9]
                                ,   l2_H_rt
                                ,   Hmap_rt
                                ,   H_sizert
                                ,   H_base_in_te_v
                                ,   Hlinc
                                ,   Hlinc_has_offset
                                ,   batch
                                ,   theta
                                ,   d_te_v
                                ,   stream[0]
                                );
            cudaDeviceSynchronize();

            cu_stat = cudaMemcpyToSymbolAsync(    list2H_rt
                                                , l2_H_rt
                                                , 2*Y_N*Y_N*sizeof(int)
                                                , 0
                                                , cudaMemcpyHostToDevice
                                                , stream[0]
                                                );
            // if (cu_stat != cudaSuccess)
            // {
            //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
            // }
            // cudaDeviceSynchronize();

            H_sizert[0] += 1;
            cu_stat = cudaMemcpyToSymbolAsync(    Hsize_rt
                                                , H_sizert
                                                , sizeof(int)
                                                , 0
                                                , cudaMemcpyHostToDevice
                                                , stream[0]
                                                );
            // if (cu_stat != cudaSuccess)
            // {
            //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
            // }
            // cudaDeviceSynchronize();

            cu_stat = cudaMemcpyToSymbolAsync(    list_Hmap_rt
                                                , Hmap_rt
                                                , Y_N*Y_N*sizeof(int)
                                                , 0
                                                , cudaMemcpyHostToDevice
                                                , stream[0]
                                                );
            // if (cu_stat != cudaSuccess)
            // {
            //    mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
            // }
            // cudaDeviceSynchronize();


            cudaDeviceSynchronize();
            
            
            sabl_mvn_get_logDetH<<<blocksPerGrid, THREADS_PER_BLOCK, 0, stream[0]>>>( d_theta
                                                                                    , d_te_v
                                                                                    , d_ret_pdf
                                                                                    , Y_N_set_constant[0]
                                                                                    , batch
                                                                                    );
            
            

            /*
             * Calculate: sum of residual^2
             */
            cudaStreamCreate(&stream[1]);

            cudaGetLastError();

            //mexPrintf("Try to set X_K_rt: %d\n", X_K_set_constant[0]);

            cu_stat = cudaMemcpyToSymbol(X_K_rt, X_K_set_constant, sizeof(int));
            if (cu_stat != cudaSuccess)
            {
                mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
            }

            /* #ifdef DEBUG
             *
                X_K_set_constant[0] = 0;
                cu_stat = cudaMemcpyFromSymbol( X_K_set_constant, X_K_rt, sizeof(int));
                if (cu_stat != cudaSuccess)
                {
                    mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
                }
                mexPrintf("Final X_K_rt: %d\n", X_K_set_constant[0]);
            *
            */

            //mexPrintf("Try to set Y_N_rt: %d\n", Y_N_set_constant[0]);

            cu_stat = cudaMemcpyToSymbol(Y_N_rt, Y_N_set_constant, sizeof(int));
            if (cu_stat != cudaSuccess)
            {
                mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
            }

            /*
             *
                Y_N_set_constant[0] = 0;
                cu_stat = cudaMemcpyFromSymbol( Y_N_set_constant, Y_N_rt, sizeof(int));
                if (cu_stat != cudaSuccess)
                {
                    mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
                }
                mexPrintf("Final Y_N_rt: %d\n", Y_N_set_constant[0]);
            *
            */


	    // 
	    //  Process matrix: A
        //  Transfer to GPU
        // - the list of non-zero elements for A: list2A_rt
        // - the quantity of such elements: Asize_rt
        // - the list of mappings for A: list_Amap_rt
        // 
	    int Amap_rt[Y_N*Y_N];
	    int A_sizert[1];	
	    int l2_A_rt[2*Y_N*Y_N];

            convert_with_linc(      prhs[3]
                                ,   l2_A_rt
                                ,   Amap_rt
                                ,   A_sizert
                                ,   A_base_in_te_v
                                ,   Alinc
                                ,   Alinc_has_offset
                                ,   batch
                                ,   theta
                                ,   d_te_v
                                ,   stream[0]
                                );
        cudaDeviceSynchronize();
        A_sizert[0] += 1;
	    cu_stat = cudaMemcpyToSymbolAsync(    Asize_rt
                                            , A_sizert
                                            , sizeof(int)
                                            , 0
                                            , cudaMemcpyHostToDevice
                                            , stream[1]
                                            );
        // if (cu_stat != cudaSuccess)
        // {
        //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        // }
        // cudaDeviceSynchronize();

        //    A_sizert[0] = 0;
        //    cu_stat = cudaMemcpyFromSymbol( A_sizert, Asize_rt, sizeof(int));
        //    if (cu_stat != cudaSuccess)
        //    {
        //        mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        //    }
        //    mexPrintf("Final Asize_rt: %d\n", A_sizert[0]);


	    // 
	    // Transfer the list of non-zero elements for A: list2A_rt
	    // 

        // cu_stat = cudaMemcpyFromSymbol( l2_A_rt, list2A_rt, 2*Y_N*Y_N*sizeof(int));
        //    if (cu_stat != cudaSuccess)
        //    {
        //       mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        //    }
        //    else
        //    {
        //        for ( uint in = 0; in < 2*Y_N*Y_N; ++in )
        //        {
        //            mexPrintf("Compiled list2A_rt: %d\n", l2_A_rt[in]);
        //        }
        //    }

	    // for ( uint in = 0; in < 2*A_sizert[0]; ++in)
	    // {
		//     l2_A_rt[in] = (int)( *(l2A_ptr+in) );
            // mexPrintf("A: %d\n", l2_A_rt[in]);
	    // }
	    cu_stat = cudaMemcpyToSymbolAsync(    list2A_rt
                                            , l2_A_rt
                                            , 2*Y_N*Y_N*sizeof(int)
                                            , 0
                                            , cudaMemcpyHostToDevice
                                            , stream[1]
                                            );
        // if (cu_stat != cudaSuccess)
        // {
        //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        // }
        // cudaDeviceSynchronize();

	    // for ( uint in = 0; in < 2*Y_N*Y_N; ++in)
	    // {
		//    l2_A_rt[in] = 0;
	    // }
        // cu_stat = cudaMemcpyFromSymbol( l2_A_rt, list2A_rt, 2*Y_N*Y_N*sizeof(int));
        //    if (cu_stat != cudaSuccess)
        //    {
        //        mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        //    }
        //    else
        //    {
        //        for ( uint in = 0; in < 2*Y_N*Y_N; ++in )
        //        {
        //            mexPrintf("Final list2A_rt: %d\n", l2_A_rt[in]);
        //        }
        //    }

	    // 
	    // Transfer the list of mappings for A: list_Amap_rt
	    // 

        //cu_stat = cudaMemcpyFromSymbol( Amap_rt, list_Amap_rt, Y_N*Y_N*sizeof(int));
        //    if (cu_stat != cudaSuccess)
        //    {
        //        mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        //    }
        //    else
        //    {
        //        for ( uint in = 0; in < Y_N*Y_N; ++in )
        //        {
        //            mexPrintf("Compiled list_Amap_rt: %d\n", Amap_rt[in]);
        //        }
        //    }

	    // for ( uint im = 0; im < A_sizert[0]; ++im)
	    // {
		//     Amap_rt[im] = (int)( *(Amap_ptr+im) );
            // mexPrintf("A map: %d\n", Amap_rt[im]);
	    // }
	    cu_stat = cudaMemcpyToSymbolAsync(    list_Amap_rt
                                            , Amap_rt
                                            , Y_N*Y_N*sizeof(int)
                                            , 0
                                            , cudaMemcpyHostToDevice
                                            , stream[1]
                                            );
        // if (cu_stat != cudaSuccess)
        // {
        //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        // }
        // cudaDeviceSynchronize();

	    // 
	    //  Process matrix: B
	    //
        // Transfer to GPU
        // - the list of non-zero elements for B: list2B_rt
        // - the quantity of such elements: Bsize_rt
        // - the list of mappings for B: list_Bmap_rt
        //

	    int Bmap_rt[X_K*X_K];
	    int B_sizert[1];	
	    int l2_B_rt[2*X_K*X_K];

            convert_with_linc(      prhs[6]
                                ,   l2_B_rt
                                ,   Bmap_rt
                                ,   B_sizert
                                ,   B_base_in_te_v
                                ,   Blinc
                                ,   Blinc_has_offset
                                ,   batch
                                ,   theta
                                ,   d_te_v
                                ,   stream[0]
                                );
        cudaDeviceSynchronize();

        B_sizert[0] += 1;
	    cu_stat = cudaMemcpyToSymbolAsync(    Bsize_rt
                                            , B_sizert
                                            , sizeof(B_sizert)
                                            , 0
                                            , cudaMemcpyHostToDevice
                                            , stream[1]
                                            );
        // if (cu_stat != cudaSuccess)
        // {
        //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        // }
        // cudaDeviceSynchronize();

	    // 
	    // Transfer the list of non-zero elements for B: list2B_rt
	    // 

	    cu_stat = cudaMemcpyToSymbolAsync(    list2B_rt
                                            , l2_B_rt
                                            , 2*X_K*X_K*sizeof(int)
                                            , 0
                                            , cudaMemcpyHostToDevice
                                            , stream[1]
                                            );
        // if (cu_stat != cudaSuccess)
        // {
        //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        // }
        // cudaDeviceSynchronize();

	    // 
	    // Transfer the list of mappings for B: list_Bmap_rt
	    // 

	    cu_stat = cudaMemcpyToSymbolAsync(    list_Bmap_rt
                                            , Bmap_rt
                                            , X_K*X_K*sizeof(int)
                                            , 0
                                            , cudaMemcpyHostToDevice
                                            , stream[1]
                                            );
        // if (cu_stat != cudaSuccess)
        // {
        //     mexErrMsgIdAndTxt(errIdCuda, errMsgCuda);
        // }
        // cudaDeviceSynchronize();

	    //
	    // OK - now launch the main multiplication kernel
	    //
            matrixdim =   mxGPUGetDimensions(mydata);
            cudaStreamCreate(&stream[2]);
            cudaDeviceSynchronize();

            for (int lh=0; lh<Y_N_set_constant[0]; ++lh)
            {
                       sabl_mvn<<<blocksPerGrid, THREADS_PER_BLOCK, 0, stream[lh%3]>>>( d_theta
								, batch
                                , lh+1
                                , lh+1
								, d_te_v
								, d_mydata
								, matrixdim[0]
								, matrixdim[1]
								, d_trsum + lh*batch
                				);
            }


            /*
             * Calculate: log(det(I-A))
             */

            sabl_mvn_getI_minusA<<<blocksPerGrid, THREADS_PER_BLOCK, 0, stream[2]>>>( d_theta, d_te_v, batch, Y_N_set_constant[0], d_dbg );

            h_A[0] = &d_dbg[0];
            for (int ib=1; ib<batch; ++ib)
            {
               h_A[ib] = h_A[0] + ib * Y_N_set_constant[0] * Y_N_set_constant[0];
            }    
            cudaMemcpyAsync(d_suB,
                        h_A,
                        sizeof(double*) * batch,
                        cudaMemcpyHostToDevice,
                        stream[2]);
            cublasSetStream(handle,
                            stream[2]);
            cublasDgetrfBatched(handle,
                                   Y_N_set_constant[0],
                                   d_suB,
                                   Y_N_set_constant[0],
                                   pivots_d,
                                   opstatus_d,
                                   batch);

            if (stat != CUBLAS_STATUS_SUCCESS)
            {
                   if (stat == CUBLAS_STATUS_NOT_INITIALIZED )
                    {
                        mexWarnMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check that cublasCreate was called.");
                    }
                   if (stat == CUBLAS_STATUS_ALLOC_FAILED )
                    {
                        mexWarnMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check that cudaMalloc has been freed.");
                    }
                   if (stat == CUBLAS_STATUS_INVALID_VALUE )
                    {
                        mexWarnMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check that paramaters are in range.");
                    }
                   if (stat == CUBLAS_STATUS_MAPPING_ERROR )
                    {
                        mexWarnMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check textures have been unbound.");
                    }
                   if (stat == CUBLAS_STATUS_EXECUTION_FAILED )
                   {
                       mexWarnMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Check why some kernel failed to launch.");
                   }
                   if (stat == CUBLAS_STATUS_INTERNAL_ERROR )
                    {
                        mexWarnMsgIdAndTxt("parallel:gpu:cublasdet:GeTrf","Cublas internal error detected.");
                    }
            }

            sabl_logdet<<<blocksPerGrid, THREADS_PER_BLOCK, 0, stream[2]>>>(batch, Y_N_set_constant[0], pivots_d, d_dbg);

            cudaDeviceSynchronize();

            /*
             * Gather and return the partial sums
             */

            sabl_mvn_gatherpdf<<<blocksPerGrid, THREADS_PER_BLOCK>>>(	  batch
									, *T_ptr 
									, Y_N_set_constant[0]
									, d_ret_pdf
									, d_dbg
									, d_trsum
									); 

            // cudaDeviceSynchronize();

            /*
             *
             * Returned values 
             *
             */

            /* Wrap the result up as a MATLAB gpuArray for return. */
            plhs[0] = mxGPUCreateMxArrayOnGPU(ret_pdf);
            
            // plhs[0] = mxGPUCreateMxArrayOnGPU(theta_extended_volatile);
            // plhs[1] = mxGPUCreateMxArrayOnGPU(dbg);
            // plhs[2] = mxGPUCreateMxArrayOnGPU(trsum);
            

            /*
             * Deallocate dynamic memory
             */

            cudaFreeHost((void**) h_A);
            cudaFree((void**) d_suB);
            cudaFree((void*) pivots_d);
            cudaFree((void*) opstatus_d);

            cudaStreamDestroy(stream[0]);
            cudaStreamDestroy(stream[1]);
            cudaStreamDestroy(stream[2]);

            mxGPUDestroyGPUArray(theta_extended_volatile);

            mxGPUDestroyGPUArray(ret_pdf);
            mxGPUDestroyGPUArray(dbg);
            mxGPUDestroyGPUArray(trsum);
        }
        /*
         * The mxGPUArray pointers are host-side structures that refer to device
         * data. These must be destroyed before leaving the MEX function.
         */
        mxGPUDestroyGPUArray(theta);
        mxGPUDestroyGPUArray(mydata);
        mxGPUDestroyGPUArray(pdf);
        mxGPUDestroyGPUArray(Alinc);
        mxGPUDestroyGPUArray(Blinc);
        mxGPUDestroyGPUArray(Hlinc);
        mxGPUDestroyGPUArray(theta_extended);
        
        mxFree((void *)matrixdim); 
        mxFree((void *)Alinc_size); 
        mxFree((void *)Blinc_size); 
        mxFree((void *)Hlinc_size); 

        cublasDestroy(handle);
    }
    else
    {   
        plhs[0] = mxCreateCharMatrixFromStrings(13, (const char **)compiled_all);
    } 
}
