/*
mex -largeArrayDims -L/usr/local/cuda-6.5/lib64 -lcublas l_negative_binomial_CUDA.cu
*/

#ifdef SABL_WIN
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef SABL_LINUX
#include "sys/time.h"
#endif

#include "mex.h"
//#include "matrix.h"
#include "gpu/mxGPUArray.h"
//#include "cuda_runtime.h" //May be unnecessary as nvcc takes care of inclusion of cuda_runtime.h. -HX, 21 July, 2014
#include "cublas_v2.h"
#include "thrust/device_ptr.h"
#include "thrust/partition.h"
#include "thrust/host_vector.h"
#include "thrust/device_vector.h"
#include "thrust/execution_policy.h"
#include "thrust/copy.h"
#include "thrust/fill.h"
#include "thrust/sequence.h"
#include "thrust/inner_product.h"


#define CUDA_CALL(x) {							\
    cudaError_t _m_cudaStat = x;					\
    if (_m_cudaStat != cudaSuccess) {					\
	fprintf(stderr, "Error %s at line %d in file %s\n",		\
	cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);		\
	exit(1);							\
    } }

#define CUBLAS_CALL(x) {					    \
    cublasStatus_t  _m_cublasStat = x;				    \
    if ( _m_cublasStat != CUBLAS_STATUS_SUCCESS ) {		    \
	fprintf(stderr, "Error %s at line %d in file %s\n",	    \
	cublasGetErrorString(_m_cublasStat), __LINE__, __FILE__);   \
	exit(1);						    \
    } }


template<typename T>
struct is_greater_than_0 
{
    __host__ __device__ 
    bool operator() (const T& x) {
	return (x > 0);
    }
};


const char* cublasGetErrorString(cublasStatus_t status)
{
    switch(status)
    {
        case CUBLAS_STATUS_SUCCESS: return "CUBLAS_STATUS_SUCCESS";
        case CUBLAS_STATUS_NOT_INITIALIZED: return "CUBLAS_STATUS_NOT_INITIALIZED";
        case CUBLAS_STATUS_ALLOC_FAILED: return "CUBLAS_STATUS_ALLOC_FAILED";
        case CUBLAS_STATUS_INVALID_VALUE: return "CUBLAS_STATUS_INVALID_VALUE"; 
        case CUBLAS_STATUS_ARCH_MISMATCH: return "CUBLAS_STATUS_ARCH_MISMATCH"; 
        case CUBLAS_STATUS_MAPPING_ERROR: return "CUBLAS_STATUS_MAPPING_ERROR";
        case CUBLAS_STATUS_EXECUTION_FAILED: return "CUBLAS_STATUS_EXECUTION_FAILED"; 
        case CUBLAS_STATUS_INTERNAL_ERROR: return "CUBLAS_STATUS_INTERNAL_ERROR"; 
    }
    return "unknown error";
}

#ifdef SABL_LINUX
double get_wall_time(){
    struct timeval time;
    if (gettimeofday(&time,NULL)){
        //  Handle error
        return 0;
    }
    return (double)time.tv_sec + (double)time.tv_usec * .000001;
}
#endif

__global__ void GPU_nbpdf(	
			double * logp, /*pointer to logp*/
			double * y, int ny, /*pointer to y, number of elements in y*/
			double * logmean, /*pointer to logmean*/
			double * logoverdispersion, int n /*pointer to logoverdispersion, number of elements in logp, logmean, and logoverdispersion*/
			) {
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int nthreads = gridDim.x * blockDim.x;
    while (tid < n) {
	double logoverdispersiontid = logoverdispersion[tid];
	double r = exp(logmean[tid] - logoverdispersiontid);
	double ytid = (ny >= n) ? y[tid] : y[0];
	double log1pv = log1p(exp(logoverdispersiontid)); /*note log1pv here = -1.0*log1p in m_nbpdf.m*/
	logp[tid] = r * (logoverdispersiontid - log1pv) - ytid * log1pv + lgamma(r+ytid) - lgamma(r) - lgamma(ytid+1.0);

	tid += nthreads;
    }
}

void mexFunction(
		int nlhs, mxArray *plhs[],
		int nrhs, const mxArray *prhs[]
		) 
{
#ifdef _DEBUG_
    double diff, start = get_wall_time();
#endif
    mxInitGPU(); 
    cublasHandle_t handle;
    CUBLAS_CALL( cublasCreate(&handle) )
    
    /*
    syntax 
    logp = l_negative_binomial(bet, gam, data, x_pointer, ...
      z_pointer, y_pointer, intensity)

    e.g.
      bet            16384x2               108  gpuArray              
      gam            16384x2               108  gpuArray              
      data             500x4               108  gpuArray              
      x_pointer          1x2                16  double                
      z_pointer          1x2                16  double                
      y_pointer          1x1                 8  double                
      intensity        500x1               108  gpuArray              

    */
	
    if	(
	    ( 7 != nrhs )
	||  !(mxIsGPUArray(prhs[0]))
	||  !(mxIsGPUArray(prhs[1]))
	||  !(mxIsGPUArray(prhs[2]))
	||  !(mxIsDouble(prhs[3]))
	||  !(mxIsDouble(prhs[4]))
	||  !(mxIsDouble(prhs[5]))
	||  !(mxIsGPUArray(prhs[6]))
	)
    { 
    	/*mexErrMsgIdAndTxt("m_negative_binomail.cu:InvalidInput", "Usage: logp = l_negative_binomial(bet, gam, data, x_pointer, z_pointer, y_pointer, intensity). Bet, gam, data, intensity should be gpuArray, x_pointer, z_pointer and y_pointer should be double array");*/
    } 
    const mxGPUArray * bet_mxGPUArray = mxGPUCreateFromMxArray(prhs[0]);
    const mxGPUArray * gam_mxGPUArray = mxGPUCreateFromMxArray(prhs[1]);
    const mxGPUArray * data_mxGPUArray = mxGPUCreateFromMxArray(prhs[2]);
    const mxGPUArray * intensity_mxGPUArray = mxGPUCreateFromMxArray(prhs[6]);
    const mxArray * x_pointer_mxArray = prhs[3];
    const mxArray * z_pointer_mxArray = prhs[4];
    const mxArray * y_pointer_mxArray = prhs[5];


    if (
	mxGPUGetClassID(bet_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(gam_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(data_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(intensity_mxGPUArray) != mxDOUBLE_CLASS
        || mxGetClassID(x_pointer_mxArray) != mxDOUBLE_CLASS
        || mxGetClassID(z_pointer_mxArray) != mxDOUBLE_CLASS
        || mxGetClassID(y_pointer_mxArray) != mxDOUBLE_CLASS
	)
    {
        /*mexErrMsgIdAndTxt("l_negative_binomial:InvalidClassID", "Input arrays (maybe  mxArray or mxGPUArray) are not double.");*/
    }


    const double constone = 1.0;
    const double constzer = 1.0;
    /*
     * Extract a pointer to the underlying arrays
     */
    const double	* bet = (double const *)(mxGPUGetDataReadOnly(bet_mxGPUArray));
    const double	* gam = (double const *)(mxGPUGetDataReadOnly(gam_mxGPUArray));
    const double       * data = (double const *)(mxGPUGetDataReadOnly(data_mxGPUArray));
    mwSize	intensity_len = mxGPUGetNumberOfElements(intensity_mxGPUArray);
    const double  * intensity = (double const *)(mxGPUGetDataReadOnly(intensity_mxGPUArray));

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    size_t	x_pointer_len = mxGetNumberOfElements(x_pointer_mxArray);
    double	  * x_pointer = (double *)(mxGetPr(x_pointer_mxArray));
    thrust::host_vector<int> x_pointer_vec(x_pointer_len);
    for (int i = 0; i<x_pointer_len; i++) {
	x_pointer_vec[i] = (int)x_pointer[i]-1; /*From 1-based to 0-based*/
    }

    size_t	z_pointer_len = mxGetNumberOfElements(z_pointer_mxArray);
    double	  * z_pointer = (double *)(mxGetPr(z_pointer_mxArray));
    thrust::host_vector<int> z_pointer_vec(z_pointer_len);
    for (int i = 0; i<z_pointer_len; i++) {
	z_pointer_vec[i] = (int)z_pointer[i]-1;
    }

    size_t	y_pointer_len = mxGetNumberOfElements(y_pointer_mxArray);
    double	  * y_pointer = (double *)(mxGetPr(y_pointer_mxArray));
    thrust::host_vector<int> y_pointer_vec(y_pointer_len);
    for (int i = 0; i<y_pointer_len; i++) {
	y_pointer_vec[i] = (int)y_pointer[i]-1;
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif



    mwSize const * dims = mxGPUGetDimensions(bet_mxGPUArray);
    size_t betM = (size_t)dims[0];
    if (dims) mxFree((void*)dims);

    dims = mxGPUGetDimensions(gam_mxGPUArray);
    if (dims) mxFree((void*)dims);

    dims = mxGPUGetDimensions(data_mxGPUArray);
    size_t dataM = (size_t)dims[0];
    if (dims) mxFree((void*)dims);

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

 
    double * intensity_true = NULL;
    CUDA_CALL( cudaMalloc( &intensity_true, intensity_len * sizeof(double) ) )
    thrust::device_ptr<double> intensity_iter(intensity_true);
    int intensity_true_len = thrust::copy_if(thrust::device, intensity, intensity+intensity_len, intensity_iter, is_greater_than_0<double>() ) - intensity_iter;

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> intensity_true_len = %d", __FILE__, __LINE__, intensity_true_len);
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    
    thrust::device_vector<double> x_vector(intensity_true_len * x_pointer_len, 0.0);
    double * x = x_vector.data().get();
    for (int i = 0; i<x_pointer_len; i++) {
	thrust::copy_if(thrust::device, data+dataM*x_pointer_vec[i], data+dataM*(x_pointer_vec[i]+1), intensity, x+intensity_true_len*i, is_greater_than_0<double>() );
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> intensity_true_len = %d", __FILE__, __LINE__, intensity_true_len);
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    thrust::device_vector<double> z_vector( intensity_true_len * z_pointer_len, 0.0); 
    double * z = z_vector.data().get();
    for (int i = 0; i<z_pointer_len; i++) {
	thrust::copy_if(thrust::device, data+dataM*z_pointer_vec[i], data+dataM*(z_pointer_vec[i]+1), intensity, z+intensity_true_len*i, is_greater_than_0<double>() );
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    thrust::device_vector<double> y_vector(intensity_true_len * y_pointer_len, 0.0);
    double * y = y_vector.data().get();
    for (int i = 0; i<y_pointer_len; i++) {
	thrust::copy_if(thrust::device, data+dataM*y_pointer_vec[i], data+dataM*(y_pointer_vec[i]+1), intensity, y+intensity_true_len*i, is_greater_than_0<double>() );
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    thrust::device_vector<double> logmean_vector(betM * intensity_true_len, 0.0);
    double * logmean = logmean_vector.data().get();

    thrust::device_vector<double>  logoverdispersion_vector(betM * intensity_true_len, 0.0);
    double * logoverdispersion = logoverdispersion_vector.data().get();

    thrust::device_vector<double> logp_overintensity_vector(betM * intensity_true_len, 0.0);
    double * logp_overintensity = logp_overintensity_vector.data().get();

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization of vectors took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    CUBLAS_CALL( cublasDgemm(handle, 
			CUBLAS_OP_N, CUBLAS_OP_T, 
			betM, intensity_true_len, x_pointer_len, 
			&constone,
			bet, betM,
			x, intensity_true_len,
			&constzer,
			logmean, betM ) )

    CUBLAS_CALL( cublasDgemm(handle, 
			CUBLAS_OP_N, CUBLAS_OP_T, 
			betM, intensity_true_len, z_pointer_len, 
			&constone,
			gam, betM,
			z, intensity_true_len,
			&constzer,
			logoverdispersion, betM ) )

#ifdef _DEBUG_
    CUDA_CALL(cudaDeviceSynchronize() )
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Matrix multiplication took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    cudaStream_t *stream = (cudaStream_t *)calloc(intensity_true_len, sizeof(cudaStream_t) );
    for (int st = 0; st < intensity_true_len; st++)
    {
	CUDA_CALL( cudaStreamCreate(&stream[st]) );
	GPU_nbpdf <<< 128, 128, 0, stream[st] >>> (logp_overintensity+st*betM, y+st, 1, logmean+st*betM, logoverdispersion+st*betM, betM);	
    }

#ifdef _DEBUG_
    CUDA_CALL(cudaDeviceSynchronize() )
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Kernel took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    for (int st = 0; st < intensity_true_len; st++)
    {
	CUDA_CALL( cudaStreamDestroy(stream[st]) );
    }

    mwSize dim_plhs = betM; 
    mxGPUArray * logp_mxGPUArray = mxGPUCreateGPUArray(1, &dim_plhs, mxGPUGetClassID(data_mxGPUArray), mxREAL, MX_GPU_INITIALIZE_VALUES); 
    double * logp = (double *)mxGPUGetData(logp_mxGPUArray);
    CUBLAS_CALL( cublasDgemm(handle, 
			CUBLAS_OP_N, CUBLAS_OP_N, 
			betM, 1, intensity_true_len, 
			&constone,
			logp_overintensity, betM,
			intensity_true, intensity_true_len,
			&constzer,
			logp, betM ) )

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Matrix multiplication 2 took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif


    plhs[0] = mxGPUCreateMxArrayOnGPU(logp_mxGPUArray); 
    if (logp_mxGPUArray) mxGPUDestroyGPUArray(logp_mxGPUArray);
    if (intensity_true) { cudaFree(intensity_true); intensity_true = NULL; }
    if (stream) { free(stream); stream = NULL; }
    if (bet_mxGPUArray) { mxGPUDestroyGPUArray(bet_mxGPUArray); bet_mxGPUArray = NULL; }
    if (gam_mxGPUArray) { mxGPUDestroyGPUArray(gam_mxGPUArray); gam_mxGPUArray = NULL; }
    if (data_mxGPUArray) { mxGPUDestroyGPUArray(data_mxGPUArray); data_mxGPUArray = NULL; }
    if (intensity_mxGPUArray) { mxGPUDestroyGPUArray(intensity_mxGPUArray); intensity_mxGPUArray = NULL; }
    CUBLAS_CALL( cublasDestroy(handle) )

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Clean-up took %f seconds", __FILE__, __LINE__, (double)diff);
#endif
}



