#ifdef SABL_WIN
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef SABL_LINUX
#include "sys/time.h"
#endif

#include <mex.h>
#include <matrix.h>
#include <math.h>
#include <gpu/mxGPUArray.h>
#include <cuda.h>
#include <cublas_v2.h>
#include "thrust/host_vector.h"
#include "thrust/device_vector.h"
#include "thrust/copy.h"
#include "thrust/fill.h"
#include "thrust/execution_policy.h"
// #include "sys/time.h"

#define LOG2PI		    1.837877066409345	    // log(2*pi)


#define CUDA_CALL(x) {							\
    cudaError_t _m_cudaStat = x;					\
    if (_m_cudaStat != cudaSuccess) {					\
	fprintf(stderr, "Error %s at line %d in file %s\n",		\
	cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);		\
	exit(1);							\
    } }

#define CUBLAS_CALL(x) {					    \
    cublasStatus_t  _m_cublasStat = x;				    \
    if ( _m_cublasStat != CUBLAS_STATUS_SUCCESS ) {		    \
	fprintf(stderr, "Error %s at line %d in file %s\n",	    \
	cublasGetErrorString(_m_cublasStat), __LINE__, __FILE__);   \
	exit(1);						    \
    } }

template<typename T>
struct is_greater_than_0 
{
    __host__ __device__ 
    bool operator() (const T& x) {
	return (x > 0);
    }
};



const char* cublasGetErrorString(cublasStatus_t status)
{
    switch(status)
    {
        case CUBLAS_STATUS_SUCCESS: return "CUBLAS_STATUS_SUCCESS";
        case CUBLAS_STATUS_NOT_INITIALIZED: return "CUBLAS_STATUS_NOT_INITIALIZED";
        case CUBLAS_STATUS_ALLOC_FAILED: return "CUBLAS_STATUS_ALLOC_FAILED";
        case CUBLAS_STATUS_INVALID_VALUE: return "CUBLAS_STATUS_INVALID_VALUE"; 
        case CUBLAS_STATUS_ARCH_MISMATCH: return "CUBLAS_STATUS_ARCH_MISMATCH"; 
        case CUBLAS_STATUS_MAPPING_ERROR: return "CUBLAS_STATUS_MAPPING_ERROR";
        case CUBLAS_STATUS_EXECUTION_FAILED: return "CUBLAS_STATUS_EXECUTION_FAILED"; 
        case CUBLAS_STATUS_INTERNAL_ERROR: return "CUBLAS_STATUS_INTERNAL_ERROR"; 
    }
    return "unknown error";
}

#ifdef SABL_LINUX
double get_wall_time(){
    struct timeval time;
    if (gettimeofday(&time,NULL)){
        //  Handle error
        return 0;
    }
    return (double)time.tv_sec + (double)time.tv_usec * .000001;
}
#endif

struct _subtract_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: y-mu
	<1>: mu
	<2>: y
	 */
	thrust::get<0>(t) = thrust::get<2>(t) - thrust::get<1>(t);
    }
};

struct _logfabs_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: logfabs
	<1>: ymu
	 */
	thrust::get<0>(t) = log(fabs(thrust::get<1>(t)));
    }
};


struct _sub_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	thrust::get<0>(t) = 2*thrust::get<1>(t) - thrust::get<2>(t);
    }
};


struct _exp_functor_
{
    template<typename T>
    __host__ __device__ 
    void operator() (T& t) {
	t = exp(t);
    }
};


struct _exp_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	thrust::get<0>(t) = exp(thrust::get<1>(t));
    }
};



struct _add_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	thrust::get<0>(t) = thrust::get<1>(t) + thrust::get<2>(t) + LOG2PI;
    }
};




struct _logsubtract_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: y-mu
	<1>: mu
	<2>: y
	 */
	thrust::get<0>(t) = log(fabs(thrust::get<2>(t) - thrust::get<1>(t)));
    }
};


struct _divlogsubtract_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: y-mu
	<1>: mu
	<2>: y
	<3>: logvary
	 */
	thrust::get<0>(t) = exp(2*log(fabs(thrust::get<2>(t) - thrust::get<1>(t))) - thrust::get<3>(t));
    }
};


struct _normal_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: logp_overintensity
	<1>: mu
	<2>: logvary
	<3>: y
	*/
        //thrust::get<0>(t) = exp(2*log(fabs(thrust::get<3>(t) - thrust::get<1>(t))) - thrust::get<2>(t)) + thrust::get<2>(t) + LOG2PI;
        thrust::get<0>(t) = (thrust::get<3>(t) - thrust::get<1>(t)) * (thrust::get<3>(t) - thrust::get<1>(t))/exp(thrust::get<2>(t)) + thrust::get<2>(t) + LOG2PI;
	/*to be continued -HX, 29 March, 2015.*/
    }
};

/*
 * Element-wise exp(), using intrinsic mathematical function __expf (faster).
 * s - number of elements.
 */
__global__
void _exp_ (double * a, long int s)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int nthreads = gridDim.x * blockDim.x;
    while (tid < s)
    {
	//a[tid] = __expf(a[tid]);
	a[tid] = exp(a[tid]);
	tid += nthreads;
    }
}






void mexFunction(int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[])
{
#ifdef _DEBUG_
    double diff, start = get_wall_time();
#endif
    mxInitGPU(); 
    cublasHandle_t handle;
    CUBLAS_CALL( cublasCreate(&handle) )
    
    /*
    syntax 
    logp = m_normal(bet, gam, data, x_pointer, ...
      z_pointer, y_pointer, intensity)

    e.g.
      bet            16384x2               108  gpuArray              
      gam            16384x2               108  gpuArray              
      data             500x4               108  gpuArray              
      x_pointer          1x2                16  double                
      z_pointer          1x2                16  double                
      y_pointer          1x1                 8  double                
      intensity        500x1               108  gpuArray              

    */
	
    /*
     * Check the number of input and whether on CPU or on GPU.
     */
    if	(
	    ( 7 != nrhs )
	||  !(mxIsGPUArray(prhs[0]))
	||  !(mxIsGPUArray(prhs[1]))
	||  !(mxIsGPUArray(prhs[2]))
	||  !(mxIsDouble(prhs[3]))
	||  !(mxIsDouble(prhs[4]))
	||  !(mxIsDouble(prhs[5]))
	||  !(mxIsGPUArray(prhs[6]))
	) 
	{
    		/*mexErrMsgIdAndTxt("l_normal_cuda.cu:InvalidInput", "Usage: logp = m_normal(bet, gam, data, x_pointer, z_pointer, y_pointer, intensity). Bet, gam, data, intensity should be gpuArray, x_pointer, z_pointer and y_pointer should be double array");*/
	}
     
    const mxGPUArray * bet_mxGPUArray = mxGPUCreateFromMxArray(prhs[0]);
    const mxGPUArray * gam_mxGPUArray = mxGPUCreateFromMxArray(prhs[1]);
    const mxGPUArray * data_mxGPUArray = mxGPUCreateFromMxArray(prhs[2]);
    const mxGPUArray * intensity_mxGPUArray = mxGPUCreateFromMxArray(prhs[6]);
    const mxArray * x_pointer_mxArray = prhs[3];
    const mxArray * z_pointer_mxArray = prhs[4];
    const mxArray * y_pointer_mxArray = prhs[5];


    /*
     * Verify the underlying class ID.
     */
    if (
	mxGPUGetClassID(bet_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(gam_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(data_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(intensity_mxGPUArray) != mxDOUBLE_CLASS
        || mxGetClassID(x_pointer_mxArray) != mxDOUBLE_CLASS
        || mxGetClassID(z_pointer_mxArray) != mxDOUBLE_CLASS
        || mxGetClassID(y_pointer_mxArray) != mxDOUBLE_CLASS
	) 
    {
        /*mexErrMsgIdAndTxt("l_normal_CUDA:InvalidClassID", "Input arrays (maybe  mxArray or mxGPUArray) are not double.");*/
    }


    const double constone = 1.0;
    const double constzer = 0.0;
    /*
     * Extract a pointer to the underlying arrays
     */
    const double	* bet = (double const *)(mxGPUGetDataReadOnly(bet_mxGPUArray));
    const double	* gam = (double const *)(mxGPUGetDataReadOnly(gam_mxGPUArray));
    const double       * data = (double const *)(mxGPUGetDataReadOnly(data_mxGPUArray));
    mwSize	intensity_len = mxGPUGetNumberOfElements(intensity_mxGPUArray);
    const double  * intensity = (double const *)(mxGPUGetDataReadOnly(intensity_mxGPUArray));

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    size_t	x_pointer_len = mxGetNumberOfElements(x_pointer_mxArray);
    double	  * x_pointer = (double *)(mxGetPr(x_pointer_mxArray));
    thrust::host_vector<int> x_pointer_vec(x_pointer_len);
    for (int i = 0; i<x_pointer_len; i++) {
	x_pointer_vec[i] = (int)x_pointer[i]-1; /*From 1-based to 0-based*/
    }

    size_t	z_pointer_len = mxGetNumberOfElements(z_pointer_mxArray);
    double	  * z_pointer = (double *)(mxGetPr(z_pointer_mxArray));
    thrust::host_vector<int> z_pointer_vec(z_pointer_len);
    for (int i = 0; i<z_pointer_len; i++) {
	z_pointer_vec[i] = (int)z_pointer[i]-1;
    }

    size_t	y_pointer_len = mxGetNumberOfElements(y_pointer_mxArray);
    double	  * y_pointer = (double *)(mxGetPr(y_pointer_mxArray));
    thrust::host_vector<int> y_pointer_vec(y_pointer_len);
    for (int i = 0; i<y_pointer_len; i++) {
	y_pointer_vec[i] = (int)y_pointer[i]-1;
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif



    mwSize const * dims = mxGPUGetDimensions(bet_mxGPUArray);
    size_t betM = (size_t)dims[0];
    if (dims) mxFree((void*)dims);

    dims = mxGPUGetDimensions(data_mxGPUArray);
    size_t dataM = (size_t)dims[0];
    if (dims) mxFree((void*)dims);

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    double * intensity_true = NULL;
    CUDA_CALL( cudaMalloc( &intensity_true, intensity_len * sizeof(double) ) )
    thrust::device_ptr<double> intensity_iter(intensity_true);
    int intensity_true_len = thrust::copy_if(thrust::device, intensity, intensity+intensity_len, intensity_iter, is_greater_than_0<double>() ) - intensity_iter;

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Matrix multiplication 2 took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif


#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    
    thrust::device_vector<double> x_vector(intensity_true_len * x_pointer_len, 0.0);
    double * x = x_vector.data().get();
    for (int i = 0; i<x_pointer_len; i++) {
	thrust::copy_if(thrust::device, data+dataM*x_pointer_vec[i], data+dataM*(x_pointer_vec[i]+1), intensity, x+intensity_true_len*i, is_greater_than_0<double>() );
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> x_pointer_len = %d", __FILE__, __LINE__, x_pointer_len);
    {
	mexPrintf("\n<%s %d> x_pointer_len = %d: ", __FILE__, __LINE__, x_pointer_len);
	for (int i = 0; i<x_pointer_len; i++) mexPrintf(" %d", x_pointer_vec[i]);
    }
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    thrust::device_vector<double> z_vector( intensity_true_len * z_pointer_len, 0.0); 
    double * z = z_vector.data().get();
    for (int i = 0; i<z_pointer_len; i++) {
	thrust::copy_if(thrust::device, data+dataM*z_pointer_vec[i], data+dataM*(z_pointer_vec[i]+1), intensity, z+intensity_true_len*i, is_greater_than_0<double>() );
    }

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> z_pointer_len = %d", __FILE__, __LINE__, z_pointer_len);
    {
	mexPrintf("\n<%s %d> z_pointer_len = %d: ", __FILE__, __LINE__, z_pointer_len);
	for (int i = 0; i<z_pointer_len; i++) mexPrintf(" %d", z_pointer_vec[i]);
    }
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    thrust::device_vector<double> y_device_vector(intensity_true_len, 0.0);
    thrust::copy_if(thrust::device, data+dataM*y_pointer_vec[0], data+dataM*(y_pointer_vec[0]+1), intensity, y_device_vector.begin(), is_greater_than_0<double>() );
    thrust::device_vector<double> y_vector(betM * intensity_true_len, 0.0);
    for (int i = 0; i < intensity_true_len; i++)
	thrust::fill_n(y_vector.begin()+betM*i, betM, y_device_vector[i]);

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    //mexEvalString("drawnow");
    start = get_wall_time();
#endif

    thrust::device_vector<double> mu_vector(betM * intensity_true_len, 0.0);
    double * mu= mu_vector.data().get();

    thrust::device_vector<double>  logvary_vector(betM * intensity_true_len, 0.0);
    double * logvary = logvary_vector.data().get();

    thrust::device_vector<double> logp_overintensity_vector(betM * intensity_true_len, 0.0);
    double * logp_overintensity = logp_overintensity_vector.data().get();

			
#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization of vectors took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    CUBLAS_CALL( cublasDgemm(handle, 
			CUBLAS_OP_N, CUBLAS_OP_T, 
			betM, intensity_true_len, x_pointer_len, 
			&constone,
			bet, betM,
			x, intensity_true_len,
			&constzer,
			mu, betM ) )

    CUBLAS_CALL( cublasDgemm(handle, 
			CUBLAS_OP_N, CUBLAS_OP_T, 
			betM, intensity_true_len, z_pointer_len, 
			&constone,
			gam, betM,
			z, intensity_true_len,
			&constzer,
			logvary, betM ) )

#ifdef _DEBUG_
    CUDA_CALL(cudaDeviceSynchronize() )
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Matrix multiplication took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    thrust::for_each(
		    thrust::make_zip_iterator(thrust::make_tuple( logp_overintensity_vector.begin(), mu_vector.begin(), logvary_vector.begin(), y_vector.begin() ) ),
		    thrust::make_zip_iterator(thrust::make_tuple( logp_overintensity_vector.end(), mu_vector.end(), logvary_vector.end(), y_vector.end() ) ),
		    _normal_element_()
		    );

#ifdef _DEBUG_
    CUDA_CALL(cudaDeviceSynchronize() )
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Kernel took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif
    
    const double minushalf = -0.5;

    mwSize dim_plhs = betM; 
    mxGPUArray * logp_mxGPUArray = mxGPUCreateGPUArray(1, &dim_plhs, mxGPUGetClassID(data_mxGPUArray), mxREAL, MX_GPU_INITIALIZE_VALUES); 
    double * logp = (double *)mxGPUGetData(logp_mxGPUArray);
    CUBLAS_CALL( cublasDgemv( handle, CUBLAS_OP_N,
			betM, intensity_true_len,
			&minushalf,
			logp_overintensity, betM,
			intensity_true, 1,
			&constzer,
			logp, 1 ) )

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Matrix multiplication 2 took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    plhs[0] = mxGPUCreateMxArrayOnGPU(logp_mxGPUArray); 
    if (logp_mxGPUArray)	{ mxGPUDestroyGPUArray(logp_mxGPUArray);	logp_mxGPUArray		= NULL; }
    if (intensity_true)		{ cudaFree(intensity_true);			intensity_true		= NULL; }
    if (bet_mxGPUArray)		{ mxGPUDestroyGPUArray(bet_mxGPUArray);		bet_mxGPUArray		= NULL; }
    if (gam_mxGPUArray)		{ mxGPUDestroyGPUArray(gam_mxGPUArray);		gam_mxGPUArray		= NULL; }
    if (data_mxGPUArray)	{ mxGPUDestroyGPUArray(data_mxGPUArray);	data_mxGPUArray		= NULL; }
    if (intensity_mxGPUArray)	{ mxGPUDestroyGPUArray(intensity_mxGPUArray);	intensity_mxGPUArray	= NULL; }
    CUBLAS_CALL( cublasDestroy(handle) )

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Clean-up took %f seconds", __FILE__, __LINE__, (double)diff);
#endif

}
