#ifdef SABL_WIN
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef SABL_LINUX
#include "sys/time.h"
#endif

/*
l_poisson_CUDA2.cu, based on m_poisson_CUDA1.cu, changes mainly on using vector to avoid manual freeing.
 */

#include <mex.h>
#include <matrix.h>
#include <math.h>
#include <gpu/mxGPUArray.h>
#include <cuda.h>
#include <cublas_v2.h>
#include "thrust/host_vector.h"
#include "thrust/device_vector.h"
#include "thrust/copy.h"
#include "thrust/fill.h"
#include "thrust/execution_policy.h"




#define CUDA_CALL(x) {							\
    cudaError_t _m_cudaStat = x;					\
    if (_m_cudaStat != cudaSuccess) {					\
	fprintf(stderr, "Error %s at line %d in file %s\n",		\
	cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);		\
	exit(1);							\
    } }

#define CUBLAS_CALL(x) {					    \
    cublasStatus_t  _m_cublasStat = x;				    \
    if ( _m_cublasStat != CUBLAS_STATUS_SUCCESS ) {		    \
	fprintf(stderr, "Error %s at line %d in file %s\n",	    \
	cublasGetErrorString(_m_cublasStat), __LINE__, __FILE__);   \
	exit(1);						    \
    } }

template<typename T>
struct is_greater_than_0 
{
    __host__ __device__ 
    bool operator() (const T& x) {
	return (x > 0);
    }
};



const char* cublasGetErrorString(cublasStatus_t status)
{
    switch(status)
    {
        case CUBLAS_STATUS_SUCCESS: return "CUBLAS_STATUS_SUCCESS";
        case CUBLAS_STATUS_NOT_INITIALIZED: return "CUBLAS_STATUS_NOT_INITIALIZED";
        case CUBLAS_STATUS_ALLOC_FAILED: return "CUBLAS_STATUS_ALLOC_FAILED";
        case CUBLAS_STATUS_INVALID_VALUE: return "CUBLAS_STATUS_INVALID_VALUE"; 
        case CUBLAS_STATUS_ARCH_MISMATCH: return "CUBLAS_STATUS_ARCH_MISMATCH"; 
        case CUBLAS_STATUS_MAPPING_ERROR: return "CUBLAS_STATUS_MAPPING_ERROR";
        case CUBLAS_STATUS_EXECUTION_FAILED: return "CUBLAS_STATUS_EXECUTION_FAILED"; 
        case CUBLAS_STATUS_INTERNAL_ERROR: return "CUBLAS_STATUS_INTERNAL_ERROR"; 
    }
    return "unknown error";
}

struct 	_gammaln_plus1_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: y_gammaln_plus1
	<1>: y
	 */
	thrust::get<0>(t) = lgamma(thrust::get<1>(t) + 1.0);
    }
};


struct _poisson_element_
{
    template <typename T>
    __host__ __device__ void operator()(T t)
    {
	/*
	<0>: logp_overintensity
	<1>: lambda
	<2>: y_gammaln
	<3>: y
	*/
        thrust::get<0>(t) = thrust::get<3>(t)*thrust::get<1>(t) - thrust::get<2>(t) - exp(thrust::get<1>(t));
    }
};

#ifdef SABL_LINUX
double get_wall_time(){
    struct timeval time;
    if (gettimeofday(&time,NULL)){
        //  Handle error
        return 0;
    }
    return (double)time.tv_sec + (double)time.tv_usec * .000001;
}
#endif


void mexFunction(int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[])
{
#ifdef _DEBUG_
    double diff, start = get_wall_time();
#endif
    mxInitGPU(); 
    cublasHandle_t handle;
    CUBLAS_CALL( cublasCreate(&handle) )
    
    /*
    syntax 
    logp = l_poisson_cuda(bet, data, x_pointer, ...
      y_pointer, intensity)

    e.g.
      bet            16384x2               108  gpuArray              
      data             500x4               108  gpuArray              
      x_pointer          1x2                16  double                
      y_pointer          1x1                 8  double                
      intensity        500x1               108  gpuArray              

    */
	
    /*
     * Check the number of input and whether on CPU or on GPU.
     */
    if	(
	    ( 5 != nrhs )
	||  !(mxIsGPUArray(prhs[0]))
	||  !(mxIsGPUArray(prhs[1]))
	||  !(mxIsDouble(prhs[2]))
	||  !(mxIsDouble(prhs[3]))
	||  !(mxIsGPUArray(prhs[4]))
	) 
	{
    	/*mexErrMsgIdAndTxt("l_poisson_cuda.cu:InvalidInput", "Usage: logp = l_poisson_cuda(bet, data, x_pointer, y_pointer, intensity). Bet, data, intensity should be gpuArray, x_pointer and y_pointer should be double array");*/
	}
     
    /* 
     * Create mxGPUArray to access helper functions on mxGPUArray
     */
    const mxGPUArray	* bet_mxGPUArray    = mxGPUCreateFromMxArray(prhs[0]);
    const mxGPUArray	* data_mxGPUArray   = mxGPUCreateFromMxArray(prhs[1]);
    const mxArray	* x_pointer_mxArray = prhs[2];
    const mxArray	* y_pointer_mxArray = prhs[3];
    const mxGPUArray	* intensity_mxGPUArray = mxGPUCreateFromMxArray(prhs[4]);


    /*
     * Verify the underlying class ID.
     */
    if (
	mxGPUGetClassID(bet_mxGPUArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(data_mxGPUArray) != mxDOUBLE_CLASS
        || mxGetClassID(x_pointer_mxArray) != mxDOUBLE_CLASS
        || mxGetClassID(y_pointer_mxArray) != mxDOUBLE_CLASS
        || mxGPUGetClassID(intensity_mxGPUArray) != mxDOUBLE_CLASS
	) 
    {
        /*mexErrMsgIdAndTxt("l_poisson_cuda:InvalidClassID", "Input arrays (maybe  mxArray or mxGPUArray) are not double.");*/
    }


    const double constone = 1.0;
    const double constzer = 0.0;
    /*
     * Extract a pointer to the underlying arrays
     */
    const double	* bet = (double const *)(mxGPUGetDataReadOnly(bet_mxGPUArray));
    const double       * data = (double const *)(mxGPUGetDataReadOnly(data_mxGPUArray));
    mwSize	intensity_len = mxGPUGetNumberOfElements(intensity_mxGPUArray);
    const double  * intensity = (double const *)(mxGPUGetDataReadOnly(intensity_mxGPUArray));

#ifdef _DEBUG2_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    size_t	x_pointer_len = mxGetNumberOfElements(x_pointer_mxArray);
    double	  * x_pointer = (double *)(mxGetPr(x_pointer_mxArray));
    thrust::host_vector<int> x_pointer_host_vector(x_pointer_len, 0);
#ifdef _DEBUG_
    mexPrintf("\n<%s %d> x_pointer_host_vector (len = %d): ", __FILE__, __LINE__, x_pointer_host_vector.size() );
#endif
    for (int i = 0; i<x_pointer_len; i++) {
	x_pointer_host_vector[i] = (int)x_pointer[i]-1; /*From 1-based to 0-based*/
#ifdef _DEBUG_
	mexPrintf(" %d", x_pointer_host_vector[i]);
#endif
    }

    size_t	y_pointer_len = mxGetNumberOfElements(y_pointer_mxArray);
    double	  * y_pointer = (double *)(mxGetPr(y_pointer_mxArray));
    thrust::host_vector<int> y_pointer_host_vector(y_pointer_len, 0);
#ifdef _DEBUG_
    mexPrintf("\n<%s %d> y_pointer_host_vector (len = %d): ", __FILE__, __LINE__, y_pointer_host_vector.size() );
#endif
    for (int i = 0; i<y_pointer_len; i++) {
	y_pointer_host_vector[i] = (int)y_pointer[i]-1;
#ifdef _DEBUG_
	mexPrintf(" %d", y_pointer_host_vector[i]);
#endif
    }

#ifdef _DEBUG2_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif



    mwSize const * dims = mxGPUGetDimensions(bet_mxGPUArray);
    size_t betM = (size_t)dims[0];
    mxFree((void*)dims);

    dims = mxGPUGetDimensions(data_mxGPUArray);
    size_t dataM = (size_t)dims[0];
#ifdef _DEBUG_
    size_t dataN = (size_t)dims[1];
#endif
    mxFree((void*)dims);

    double * intensity_true = NULL;
    CUDA_CALL( cudaMalloc( &intensity_true, intensity_len * sizeof(double) ) )
    thrust::device_ptr<double> intensity_iter(intensity_true);
    int intensity_true_len = thrust::copy_if(thrust::device, intensity, intensity+intensity_len, intensity_iter, is_greater_than_0<double>() ) - intensity_iter; 
    
    
    thrust::device_vector<double> x_device_vector(intensity_true_len * x_pointer_len, 0.0);
    double * x = x_device_vector.data().get();
    for (int i = 0; i<x_pointer_len; i++) {
	thrust::copy_if(thrust::device, data+dataM*x_pointer_host_vector[i], data+dataM*(x_pointer_host_vector[i]+1), intensity, x_device_vector.begin()+intensity_true_len*i, is_greater_than_0<double>() );
    }

#ifdef _DEBUG_
    double * intensity_true_host = (double *)calloc(intensity_true_len, sizeof(double) );
    CUDA_CALL( cudaMemcpy( intensity_true_host, intensity_true, intensity_true_len * sizeof(double), cudaMemcpyDeviceToHost ) )
    mexPrintf("\n<%s %d> intensity_true_host (len = %d): ", __FILE__, __LINE__, intensity_true_len);
    for (int i = 0; i<intensity_true_len; i++) {
	mexPrintf(" %f", intensity_true_host[i]);
    }
    free(intensity_true_host);

    mexPrintf("\n<%s %d> x_pointer_host_vector (len = %d): ", __FILE__, __LINE__, x_pointer_host_vector.size() );
    for (int i = 0; i<x_pointer_len; i++) {
	mexPrintf(" %d", x_pointer_host_vector[i]);
    }

    mexPrintf("\n<%s %d> y_pointer_host_vector (len = %d): ", __FILE__, __LINE__, y_pointer_host_vector.size() );
    for (int i = 0; i<y_pointer_len; i++) {
	mexPrintf(" %d", y_pointer_host_vector[i]);
    }

    double * data_host = (double *)calloc(dataM*dataN,sizeof(double));
    CUDA_CALL( cudaMemcpy(data_host, data, dataM*dataN*sizeof(double), cudaMemcpyDeviceToHost) )
    mexPrintf("\n<%s %d> data_host: ", __FILE__, __LINE__ );
    for (int i = 0; i<dataM*dataN; i++) {
	mexPrintf(" %f", data_host[i]);
    }
    free(data_host);

    thrust::host_vector<double> x_host_vector = x_device_vector;
    mexPrintf("\n<%s %d> x_host_vector (len = %d, intensity_true_len*x_pointer_len = %d: ", __FILE__, __LINE__, x_host_vector.size(), intensity_true_len*x_pointer_len);
    for (int i = 0; i<intensity_true_len*x_pointer_len; i++) {
	mexPrintf(" %f", x_host_vector[i]);
    }
    
    mexEvalString("drawnow");
    
#ifdef _DEBUG2_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif
#endif

    thrust::device_vector<double> y_intensity_true_device_vector(intensity_true_len, 0.0);
    thrust::copy_if(thrust::device, data+dataM*y_pointer_host_vector[0], data+dataM*(y_pointer_host_vector[0]+1), intensity, y_intensity_true_device_vector.begin(), is_greater_than_0<double>() );
    thrust::device_vector<double> y_gammaln_intensity_true_device_vector(intensity_true_len, 0.0);
    thrust::for_each(
	thrust::make_zip_iterator( thrust::make_tuple( y_gammaln_intensity_true_device_vector.begin(), y_intensity_true_device_vector.begin() ) ),
	thrust::make_zip_iterator( thrust::make_tuple( y_gammaln_intensity_true_device_vector.end(),   y_intensity_true_device_vector.end() ) ),
	_gammaln_plus1_()
		    );
    thrust::device_vector<double> y_allparticles_device_vector(betM * intensity_true_len, 0.0);
    thrust::device_vector<double> y_gammaln_allparticles_device_vector(betM * intensity_true_len, 0.0);
    for (int i = 0; i < intensity_true_len; i++) {
	thrust::fill_n(         y_allparticles_device_vector.begin()+betM*i, betM, y_intensity_true_device_vector[i]         );
	thrust::fill_n( y_gammaln_allparticles_device_vector.begin()+betM*i, betM, y_gammaln_intensity_true_device_vector[i]);
    }

#ifdef _DEBUG_
    thrust::host_vector<double> y_oneparticle_host_vector = y_intensity_true_device_vector;
    mexPrintf("\n<%s %d> y_oneparticle_host_vector (len = %d, intensity_true_len = %d: ", __FILE__, __LINE__, y_oneparticle_host_vector.size(), intensity_true_len);
    for (int i = 0; i<intensity_true_len; i++) {
	mexPrintf(" %f", y_oneparticle_host_vector[i]);
    }

    thrust::host_vector<double> y_allparticles_host_vector = y_allparticles_device_vector;
    mexPrintf("\n<%s %d> y_allparticles_host_vector (len = %d, betM*intensity_true_len = %d: ", __FILE__, __LINE__, y_allparticles_host_vector.size(), betM*intensity_true_len);
    for (int i = 0; i<betM*intensity_true_len; i++) {
	mexPrintf(" %f", y_allparticles_host_vector[i]);
    }

    thrust::host_vector<double> y_gammaln_allparticles_host_vector = y_gammaln_allparticles_device_vector;
    mexPrintf("\n<%s %d> y_gammaln_allparticles_host_vector (len = %d, betM*intensity_true_len = %d: ", __FILE__, __LINE__, y_gammaln_allparticles_host_vector.size(), betM*intensity_true_len);
    for (int i = 0; i<betM*intensity_true_len; i++) {
	mexPrintf(" %f", y_gammaln_allparticles_host_vector[i]);
    }
#ifdef _DEBUG2_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Initialization took %f seconds", __FILE__, __LINE__, (double)diff);
    mexEvalString("drawnow");
    start = get_wall_time();
#endif
#endif

    thrust::device_vector<double> loglambda_vector(betM * intensity_true_len, 0.0);
    thrust::device_vector<double> logp_overintensity_vector(betM * intensity_true_len, 0.0);

			
    CUBLAS_CALL( cublasDgemm(handle, 
			CUBLAS_OP_N, CUBLAS_OP_T, 
			betM, intensity_true_len, x_pointer_len, 
			&constone,
			bet, betM,
			x, intensity_true_len,
			&constzer,
			loglambda_vector.data().get(), betM ) )

#ifdef _DEBUG_
    thrust::host_vector<double> loglambda_host_vector = loglambda_vector;
    mexPrintf("\n<%s %d> loglambda_host_vector (len = %d): ", __FILE__, __LINE__, loglambda_host_vector.size() );
    for (int i = 0; i<loglambda_host_vector.size(); i++) mexPrintf(" %f", loglambda_host_vector[i]);
#endif

    thrust::for_each(
		    thrust::make_zip_iterator(thrust::make_tuple( logp_overintensity_vector.begin(), loglambda_vector.begin(), y_gammaln_allparticles_device_vector.begin(), y_allparticles_device_vector.begin() ) ),
		    thrust::make_zip_iterator(thrust::make_tuple( logp_overintensity_vector.end(),   loglambda_vector.end(),   y_gammaln_allparticles_device_vector.end(),   y_allparticles_device_vector.end() ) ),
		    _poisson_element_()
		    );

#ifdef _DEBUG_
    thrust::host_vector<double> logp_overintensity_host_vector = logp_overintensity_vector;
    mexPrintf("\n<%s %d> logp_overintensity_host_vector (len = %d): ", __FILE__, __LINE__, logp_overintensity_host_vector.size() );
    for (int i = 0; i<logp_overintensity_host_vector.size(); i++) mexPrintf(" %f", logp_overintensity_host_vector[i]);
#endif

#ifdef _DEBUG_
    CUDA_CALL(cudaDeviceSynchronize() )
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Kernel took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif
    
    mwSize dim_plhs = betM; 
    mxGPUArray * logp_mxGPUArray = mxGPUCreateGPUArray(1, &dim_plhs, mxGPUGetClassID(data_mxGPUArray), mxREAL, MX_GPU_INITIALIZE_VALUES); 
    double * logp = (double *)mxGPUGetData(logp_mxGPUArray);
    CUBLAS_CALL( cublasDgemv(handle, CUBLAS_OP_N, 
			betM, intensity_true_len, 
			&constone,
			logp_overintensity_vector.data().get(), betM,
			intensity_true, 1, 
			&constzer,
			logp, 1 ) );


#ifdef _DEBUG2_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Matrix multiplication 2 took %f seconds", __FILE__, __LINE__, (double)diff);
    start = get_wall_time();
#endif

    plhs[0] = mxGPUCreateMxArrayOnGPU(logp_mxGPUArray); 
    if (logp_mxGPUArray)	{ mxGPUDestroyGPUArray(logp_mxGPUArray); }
    if (intensity_true)		{ cudaFree(intensity_true); intensity_true = NULL; }
    if (bet_mxGPUArray)		{ mxGPUDestroyGPUArray(bet_mxGPUArray); bet_mxGPUArray = NULL; }
    if (data_mxGPUArray)	{ mxGPUDestroyGPUArray(data_mxGPUArray); data_mxGPUArray = NULL; }
    if (intensity_mxGPUArray)	{ mxGPUDestroyGPUArray(intensity_mxGPUArray); intensity_mxGPUArray = NULL; }
    CUBLAS_CALL( cublasDestroy(handle) )

#ifdef _DEBUG_
    diff = get_wall_time() - start;
    mexPrintf("\n<%s %d> Clean-up took %f seconds", __FILE__, __LINE__, (double)diff);
#endif

}
