function [logpy, v, eps] = l_egarch(muY, sigY, alph, bet, gam, ...
    p, mu, sig, y, intensity, vlag_in, epslag_in)
%  Evaluate the log predictive p.d.f. of the EGARCH model of Durham
%  and Geweke (2013) for multiple values of the parameters.
%
%  Inputs:
%  muY        Mean of y_t (n x 1)
%  sigY       Standard deviation of y_t (n x 1)
%  alph       Alpha parameters of volatility dynamics (n x K)
%  bet        Beta parmaeters of volatility dynamics (n x K)
%  gam        Gamma parameters of volatility dynamics (n x K)
%  p          Parameters p_i of epsilon normal mixture (n x I)
%  mu         Parameters mu_i of epsilon normal mixture (n x I)
%  sig        Parameters sig_i of epsilon normal mxiture (n x I)
%  y          Observed time series y_1:t2
%  intensity  Intensity vector
%  vlag_in    v_t1-1.  By convention vlag = 0 if t1 = 1.  (n x K)
%  epslag_in  epsilon_t1-1.  By convention epslag = 0 if t1 = 1.  (n x 1)
%
%  Outputs:
%  logpy      log p(y_t1:t2 | y_1:t1-1, parameters)  (n x 1)
%  v          v_t2  (n x K)
%  eps        epsilon_t2  (n x 1)

[n, K] = size(alph);
I = size(p,2);
shift = sqrt(2/pi);

logpy = zeros(n,1);
tt = find(intensity > 0);
vlag = vlag_in;
epslag = epslag_in;
for i = 1:length(tt);
    t = tt(i);
    v = alph.*vlag + bet.*(repmat(abs(epslag),1,K) - shift) + ...
        gam.*repmat(epslag,1,K);
    h = sigY.*exp(0.5*sum(v,2));
    eps = (y(t) - muY)./h;
    logpyt = log(sum(p.*normpdf(repmat(eps,1,I),mu,sig),2)./h);
    logpy = logpy + logpyt*intensity(t);
    vlag = v;
    epslag = eps;
end

logpy(isnan(logpy)) = -inf;

end