function logp = ...
    l_mvnormal(theta, F, T, Amap, Alinc, Aname, Bmap, Blinc, Bname, ...
    Hmap, Hlinc, Hname, A_tri)
%  Evaluate the log-likelihood of the MVN model at specified particles.
%  The model is
%     y_t = A*y_t + B*x_t + e_t,  e_t ~IIDN(0, Sigma)
%  where y_t is n x 1 and x_t is k x 1.
%  Equivalently if Y' = [y_1, ... , y_T] and X' = (x_1, ... , x_T] then
%     Y = Y*A' + X*B' + E, vec(E) ~ N(0, kron(Sigma, eye(T))
%
%  Inputs:
%  theta       Matrix of particles (C.JNwork x C.parameters)
%  F           Matrix F: F'F = [Y', X']*diag(intensity)*[X; Y] (n+k x n+k)
%  T           Sum of the elements in current intensity vector
%  Amap        Amap(i,j) points to the column of theta with A(i,j);
%              = 0 if identically 0  (n x n)
%  Bmap        Bmap(i,j) points to the column of theta with B(i,j);
%              = 0 if identically 0  (n x k)
%  Hmap        Hmap(i,j) points to the column of theta with H(i,j) if i~=j;
%              with log(H(i,i)) if i == j; and = 0 if identically 0 (n x n)
%              where H: H'*H = inv(Sigma).
%  A_tri       Indicator that all non-zero elements of A are strictly above
%              of strictly below the diagonal of A
%
%  Output:
%  logp        Log-likelihood evaluated at the particles (C.JNwork x 1)

%  Programming note: The CUDA version of this function will need to use
%  u_collect_CUDA. (Since p_thetapar is handled in m_message, all map 
%  entries < size(theta, 2) when l_mvnormal_CUDA is invoked.
%  (As always the computation in *_CUDA is organized differently because
%  each row of theta is a thread.)

[n, k] = size(Bmap);

%  Compute U = F*[I_n-A, -B]*H' in matrix-particle format
%  [I_n-A'; -B'] -> [U, Umap]:
[A, Amap] = u_collectpart(theta, Amap, Alinc, Aname);
[B, Bmap] = u_collectpart(theta, Bmap, Blinc, Bname);
[U, Umap] = u_adjoinpartpart(A, Amap', B, Bmap', 'columns');
[U, Umap] = u_sumfixpart([eye(n); zeros(k, n)], U, Umap, -1);

%  H' -> [H1, H1map]:
[H1, H1map] = u_collectpart(theta, Hmap', Hlinc, Hname);
sumlogH1diag = sum(H1(:,diag(H1map)), 2);  %  Used at end
H1(:,diag(H1map)) = exp(H1(:,diag(H1map)));
shiftH1map = (size(U, 2)+H1map).*(H1map>0);

%  [I_n-A'; -B']*H' -> [U, Umap]:
[U, Umap] = u_prodpartpart([U, H1], Umap, shiftH1map);

%  F*[I_n-A'; -B']*H' -> [U, Umap]:
[U, ~] = u_prodfixpart(F, U, Umap);

%  Compute log-likelihood function at all particles:
logp = -0.5*(n*T*log(2*pi) + sum(U.^2, 2)) + T*sumlogH1diag;
if A_tri
    for i = 1:n
        if Amap(i,i) > 0    
            logp = logp + T*log(abs(1-A(:,A(:,Amap(i,i)))));
        end
    end
else

A_in_cells = mat2cell([zeros(size(A,1),1) A], ...
    ones(1,size(A,1)), ...
    1+size(A,2));

A_IminusA = ...
    cellfun(@(x)(eye(size(Amap,1))-x(Amap+1)), ...
    A_in_cells, ...
    'UniformOutput', false);

A_ln_abs_det = ...
    cellfun(@(x)(log(abs(det(x)))), ...
    A_IminusA, ...
    'UniformOutput', true);

logp = logp + T.*A_ln_abs_det;
end

end