function logp = l_nbpdf(y, logmean, logoverdispersion)
%  Evaluate the Negative Binomial model log likelihood function
%  parameterized by its mean and over-dispersion
%
%  Inputs:
%  y                    The number of extra trials one must perform in 
%                       order to observe a given number R of successes has  
%                       a negative binomial distribution. (n x 1 or
%                       scalar).
%  logmean              log(r*p/(1-p)), where r is the number of successes 
%                       and p is the probability of success in a single 
%                       trial. (n x 1)
%  logoverdispersion    log(p/(1-p)), the definition of p is same as above.
%                       (n x 1)
%
%  Output:
%  logp                 Evaluation of the log likelihood (n x 1)

%  Programming note: The code below implicitly uses the Matlab version
%  of the negative binomial "p" parameter, which is equivalent to 1-p
%  in the more conventional version (e.g. Wikipedia). This coforms with
%  the invocation of nbinrnd in m_negative_binomialsim.  These conventions
%  are of no relevance or condern to the user because the negative binomial
%  SABL parameterizes this distribution by its moments.

r = exp(logmean - logoverdispersion);

log1p = - log(1 + exp(logoverdispersion));
logp = gammaln(r+y) - gammaln(r) - gammaln(y+1) + ...
    r.*(logoverdispersion + log1p) + y.*log1p;

end