function y = l_normalsim(theta, betaindex, alphaindex, ...
    data, xindex, zindex, method)
%  Simulate from the normal model using alternative values of the
%  parameters and the covariates.
%
%  Inputs:
%  theta       Each row of theta contains a particle (JN x k)
%  betaindex   Pointer to the columns of theta containing beta (k_x x 1)
%  alphaindex  Pointer to the columns of theta containing alpha (k_z x 1) 
%  data        data(t, xindex) = x_t', data(t, zindex = y_t)
%  xindex      Column indices of the covariates x in data (k_x x 1)
%  zindex      Column indices of the covariates z in data (k_z x 1)
%  method      = 1 for brute force (debug mode only); = 2 for C/CUDA 
%              mock-up; = 3 for Matlab code
%
%  Output:
%  y       y(j,t) is one simulation from the normal model using the
%          parameters in theta(j,:) and the covariate vector x(t,:) 
%                                                                  (JN x T)

T = size(data, 1);
JN = size(theta, 1);
y = u_allocate([JN, T]);

%  Brute force code
if method == 1
    for j = 1:JN
        for t = 1:T
            mu = theta(j,betaindex)*data(t,xindex)';
            sig = exp(0.5*theta(j,alphaindex)*data(t,zindex)');
            y(j,t) = mu + sig*randn;
        end
    end
    
%  C/CUDA code
elseif method == 2
    for j = 1:JN
        mu = data(:,xindex)*theta(j,betaindex)';  % T x 1
        sig = exp(0.5*data(:,zindex)*theta(j,alphaindex)');  % T x 1
        y(j,:) = (mu + sig.*randn(T,1))';
    end
    
%  Matlab code
else
    for t = 1:T
        mu = theta(:,betaindex)*data(t,xindex)';  % JN x 1
        sig = exp(0.5*theta(:,alphaindex)*data(t,zindex)');  % JN x 1
        y(:,t) = mu + sig.*randn(JN, 1);
    end
end

end     