function outcompile = compile_cuda()
%COMPILE_CUDA : MVN model helper script to compile CUDA code to MEX
%
% Global dependencies:
%
% <a href="matlab: help('MVN','M.Hmap')">M.Hmap</a>
% <a href="matlab: help('MVN','M.Amap')">M.Amap</a>
% <a href="matlab: help('MVN','M.Bmap')">M.Bmap</a>
%
% <a href="matlab: help('MVN','M.x_pointer')">M.x_pointer</a>
% <a href="matlab: help('MVN','M.y_pointer')">M.y_pointer</a>
% 
% Input(s):
%
% None - but see Global Dependencies
%
% size(M.x_pointer) defined to be K of the MVN model
% size(M.y_pointer) defined to be N in the MVN model
%
% M.x_pointer and M.y_pointer have the same meaning as in the .m file
% implementation of MVN ie. they are the indexes for columns in 
% <a href="matlab: help('MVN','M.data')">M.data</a>
%
% Output(s):
%
% l_mvnormal_CUDA.mex - for evaluating log-likelihood of the MVN model on
% GPU

global M P E
function_name = 'l_mvnormal_CUDA';
% compilenow=false leaves auto-recompilation enabled for
% detecting if the requested K or N exceed what was already
% compiled for in the previous session.  
compilenow = false;

default_K = 16;
default_N = 16;

if exist('M')
    if isfield(M,'modeldir')
        if iscell(M.modeldir)
            model_includepath = [M.modeldir{end},filesep,'cufiles'];
        else
            model_includepath = [M.modeldir,filesep,'cufiles'];
        end
    else
        assumed_location=pwd;
        model_includepath = [assumed_location,filesep,'cufiles'];
    end
else
    assumed_location=pwd;
    model_includepath = [assumed_location,filesep,'cufiles'];
end

if exist('E')
    if isfield(E,'SABLrootdir')
        if ~isempty(E.SABLrootdir)
            libpath = [E.SABLrootdir,filesep,'library',filesep,'cufiles'];
        else
            assumed_location = pwd;
            libpath = [assumed_location,filesep,'cufiles'];
        end
    else
        assumed_location = pwd;
        libpath = [assumed_location,filesep,'cufiles'];
    end
else
    libpath = model_includepath;
end

% srcfile = [model_includepath,'\',function_name,'.cu'];
srcfile = [libpath,filesep,function_name,'.cu'];

if ispc
    mxoptsfile = [model_includepath,filesep,'mex_CUDA_win64.xml'];
end
if isunix
    mxoptsfile = [model_includepath,filesep,'mex_CUDA_glnxa64.xml'];
end

if exist('P')
    if isfield(P,'projectdir')
        if ~isempty(P.projectdir)
            if iscell(P.projectdir)
                project_includepath = [P.projectdir{end},filesep,'cufiles'];
            else
                project_includepath = [P.projectdir,filesep,'cufiles'];
            end
        else
            assumed_location=pwd;
            project_includepath = [assumed_location,filesep,'cufiles'];            
        end
    else
        assumed_location=pwd;
        project_includepath = [assumed_location,filesep,'cufiles'];
    end
else
    project_includepath = model_includepath;
end



    x_map = [];
    x_list = [];

    y_map = [];
    y_list = [];

    A_map = [];
    A_list = [];

    B_map = [];
    B_list = [];

    H_map = [];
    H_list = [];

    fullstr = [];

    if exist('M')
        if isfield(M,'Amap')
            
            % Generate the list of non-zero elements in A
            if isunix
                A_list = [sprintf('%d _c_ %d', M.Amap(1,1), M.Amap(1,2))];
            elseif ispc
                A_list = [sprintf('%d,%d', M.Amap(1,1), M.Amap(1,2))];
            end
            for i=2:size(M.Amap,1)
                if isunix
                    A_list = [A_list,' _c_ %d _c_ %d', M.Amap(i,1), M.Amap(i,2)]; 
                elseif ispc
                    A_list = [A_list,sprintf(',%d,%d', M.Amap(i,1), M.Amap(i,2))];
                end
            end
            
            % For each non-zero element,  generate the mapping transform from theta to A

            if isunix
                sel = ' _t_ ';
            elseif ispc
                sel = 'c_t<';
            end
            A_map=[A_map, sel, sprintf('%d', 1)];
            
            for i=2:size(M.Amap,1) 
                if isunix
                    sel = ' _t_ ';
                elseif ispc
                    sel = 'c_t<';
                end
                    
                if isunix
                    A_map = [A_map, ' _e_ _c_ '];
                elseif ispc
                    A_map = [A_map,'>,'];                
                end
                A_map=[A_map, sel, sprintf('%d', i)];
            end
            if isunix
                A_map = [A_map, ' _e_ '];
            elseif ispc
                A_map = [A_map,'>'];
            end
            if isunix
                fullstr = [fullstr,' -DLIST2_A=''',A_list,'''', ' -DA_MAP=''',A_map,''''];       
            elseif ispc
                fullstr = [fullstr,' -DLIST2_A=''\"',A_list,'\"''', ' -DA_MAP=''\"',A_map,'\"'''];       
            end            
         end
         if isfield(M,'Bmap')
            
            % Generate the list of non-zero elements in B
            if isunix
                B_list = [sprintf('%d _c_ %d', M.Bmap(1,1), M.Bmap(1,2))];
            elseif ispc
                B_list = [sprintf('%d,%d', M.Bmap(1,1), M.Bmap(1,2))];
            end
            for i=2:size(M.Bmap,1)
                if isunix
                    B_list = [B_list,' _c_ %d _c_ %d', M.Bmap(i,1), M.Bmap(i,2)]; 
                elseif ispc
                    B_list = [B_list,sprintf(',%d,%d', M.Bmap(i,1), M.Bmap(i,2))];
                end
            end
            
            % For each non-zero element,  generate the mapping transform from theta to B

            if isunix
                sel = ' _t_ ';
            elseif ispc
                sel = 'c_t<';
            end
            B_map=[B_map, sel, sprintf('%d', size(M.Amap,1) + 1)];
            
            for i=2:size(M.Bmap,1) 
                if isunix
                    sel = ' _t_ ';
                elseif ispc
                    sel = 'c_t<';
                end
                    
                if isunix
                    B_map = [B_map, ' _e_ _c_ '];
                elseif ispc
                    B_map = [B_map,'>,'];                
                end
                B_map=[B_map, sel, sprintf('%d', size(M.Amap,1) + i)];
            end
            if isunix
                B_map = [B_map, ' _e_ '];
            elseif ispc
                B_map = [B_map,'>'];
            end
            if isunix
                fullstr = [fullstr,' -DLIST2_B=''',B_list,'''', ' -DB_MAP=''',B_map,''''];       
            elseif ispc
                fullstr = [fullstr,' -DLIST2_B=''\"',B_list,'\"''', ' -DB_MAP=''\"',B_map,'\"'''];       
            end   
         end

        if isfield(M,'Hmap')
            
            % Generate the list of non-zero elements in H
            if isunix
                H_list = [sprintf('%d _c_ %d', M.Hmap(1,1), M.Hmap(1,2))];
            elseif ispc
                H_list = [sprintf('%d,%d', M.Hmap(1,1), M.Hmap(1,2))];
            end
            for i=2:size(M.Hmap,1)
                if isunix
                    H_list = [H_list,sprintf(' _c_ %d _c_ %d', M.Hmap(i,1), M.Hmap(i,2))]; 
                elseif ispc
                    H_list = [H_list,sprintf(',%d,%d', M.Hmap(i,1), M.Hmap(i,2))];
                end
            end
            
            % For each non-zero element,  generate the mapping transform from theta to H
            if (M.Hmap(1,1) == M.Hmap(1,2)) 
                if isunix
                    sel=' _te_ ';
                elseif ispc
                    sel='c_te<';
                end
            else
                if isunix
                    sel = ' _t_ ';
                elseif ispc
                    sel = 'c_t<';
                end
            end
            H_map=[H_map, sel, sprintf('%d', size(M.Amap,1) + size(M.Bmap,1) + M.Hmap(1,2)*( M.Hmap(1,2)-1 )/2 + M.Hmap(1,1))];
            for i=2:size(M.Hmap,1) 
                if (M.Hmap(i,1) == M.Hmap(i,2)) 
                    if isunix
                        sel=' _te_ ';
                    elseif ispc
                        sel='c_te<';
                    end
                else
                    if isunix
                        sel = ' _t_ ';
                    elseif ispc
                        sel = 'c_t<';
                    end
                end    
                if isunix
                    H_map = [H_map, ' _e_ _c_ '];
                elseif ispc
                    H_map = [H_map,'>,'];                
                end
                H_map = [H_map, sel, sprintf('%d', size(M.Amap,1) + size(M.Bmap,1) + M.Hmap(i,2)*( M.Hmap(i,2)-1 )/2 + M.Hmap(i,1) )];
            end
            if isunix
                H_map = [H_map, ' _e_ '];
            elseif ispc
                H_map = [H_map,'>'];
            end
            if isunix
                fullstr = [fullstr,' -DLIST2_H=''',H_list,'''', ' -DH_MAP=''',H_map,''''];       
            elseif ispc
                fullstr = [fullstr,' -DLIST2_H=''\"',H_list,'\"''', ' -DH_MAP=''\"',H_map,'\"'''];       
            end   
        end
    end    
    
% Remove the comments to
% use the hardcoded CUDA maps of A, B and H 
    fullstr=[];
    A_list=['LIST2_A'];
    A_map=['A_MAP'];
    B_list=['LIST2_B'];
    B_map=['B_MAP'];
    H_list=['LIST2_H'];
    H_map=['H_MAP'];


    
if exist(srcfile,'file')
    if ( exist(function_name)==3 )
        % MEX was compiled previously, but we must check that M.x_pointer,
        % y-pointer; and the mapping for A , B and H still match the
        % current requirements
        %
        eval(['myans = ', function_name,';']);
        if ( strcmp(deblank(myans(7,:)), '(Y_MAP)') == 1 )
            check_ymap = [];
        else
            eval(sprintf('check_ymap = [%s];', strrep(myans(7,2:(length(deblank(myans(7,:)))-1)),'>,c_d<',',') ) );
        end
        
        if length(check_ymap) >= length(M.y_pointer)
%             if ~(check_ymap == M.y_pointer)
%                compilenow = true;
%             end
        else
            if length(M.y_pointer) > default_N
                compilenow = true;
            end
        end
        
        if ( strcmp(deblank(myans(5,:)), '(X_MAP)') == 1 )
            check_xmap = [];
        else
            eval(sprintf('check_xmap = [%s];', strrep(myans(5,2:(length(deblank(myans(5,:)))-1)),'>,c_d<',',') ) );
        end
        
        if length(check_xmap) >= length(M.x_pointer)
%             if ~(check_xmap == M.x_pointer)
%                 compilenow = true;
%             end
        else
            if length(M.x_pointer) > default_K
                compilenow = true;
            end
        end
        
        if ~strcmp( strrep( deblank(myans(8,:)), ' ' , '' ) , ['(',A_list,')'] )
            compilenow = true;
        end
        
        if ~strcmp( strrep( deblank(myans(9,:)), ' ' , '' ) , ['(',A_map,')'] )
            compilenow = true;
        end      
        
        if ~strcmp( strrep( deblank(myans(10,:)), ' ' , '' ) , ['(',B_list,')'] )
            compilenow = true;
        end
        
        if ~strcmp( strrep( deblank(myans(11,:)), ' ' , '' ) , ['(',B_map,')'] )
            compilenow = true;
        end
        
        if ~strcmp( strrep( deblank(myans(12,:)), ' ' , '' ) , ['(',H_list,')'] )
            compilenow = true;
        end
        
        if ~strcmp( strrep( deblank(myans(13,:)), ' ' , '' ) , ['(',H_map,')'] )
            compilenow = true;
        end      
        
        outcompile = compilenow;
    else
        % MEX has never been compiled
        compilenow = true;       
        outcompile = 2;
    end
    
  
    
    if compilenow == true           
        if ~isempty(M.x_pointer)
            x_list = [sprintf('%d',1)];
            for i=2:length(M.x_pointer)
                if isunix
                    x_list = [x_list,' _s_ ',sprintf('%d',i)]; 
                elseif ispc
                    x_list = [x_list,',1,',sprintf('%d',i)]; 
                end
            end
            
            x_map = [sprintf('%d',M.x_pointer(1))];
            for i=2:length(M.x_pointer)
                if isunix
                    x_map = [x_map,' _m_ ',sprintf('%d',M.x_pointer(i))];
                elseif ispc
                    x_map = [x_map,'>,c_d<',sprintf('%d',M.x_pointer(i))];
                end
            end
            

            if (length(M.x_pointer)>default_K)
                x_k = [sprintf('%d',length(M.x_pointer))];
            else
                x_k = [sprintf('%d',default_K)];
            end
            
            if isunix
                fullstr = [fullstr,' -DTHREADS_PER_BLOCK=256 -DLIST2_X=''',x_list,'''', ' -DX_MAP=''',x_map,'''', ' -DX_K=''',x_k,''''];
            elseif ispc
                fullstr = [fullstr,' -DTHREADS_PER_BLOCK=256 -DLIST2_X=''\"',x_list,'\"''', ' -DX_MAP=''\"',x_map,'\"''', ' -DX_K=''',x_k,''''];
            end
        end
        
            
        if ~isempty(M.y_pointer)
            y_list = [sprintf('%d',1)];
            for i=2:length(M.y_pointer)
                if isunix
                    y_list = [y_list,' _s_ ',sprintf('%d',i)]; 
                elseif ispc
                    y_list = [y_list,',1,',sprintf('%d',i)]; 
                end

            end
            
            y_map = [sprintf('%d',M.y_pointer(1))];
            for i=2:length(M.y_pointer)
                if isunix
                    y_map = [y_map,' _m_ ',sprintf('%d',M.y_pointer(i))]; 
                elseif ispc
                    y_map = [y_map,'>,c_d<',sprintf('%d',M.y_pointer(i))]; 
                end
            end
            
            if (length(M.y_pointer)>default_N)
                y_n = [sprintf('%d',length(M.y_pointer))];
            else
                y_n = [sprintf('%d',default_N)];
            end

            if isunix
                fullstr = [fullstr,' -DLIST2_Y=''',y_list,'''', ' -DY_MAP=''',y_map,'''', ' -DY_N=''',y_n,''''];       
            elseif ispc
                fullstr = [fullstr,' -DLIST2_Y=''\"',y_list,'\"''', ' -DY_MAP=''\"',y_map,'\"''', ' -DY_N=''',y_n,''''];       
            end
        end
        g=gpuDevice;
        if isunix
            cublaspath=[matlabroot,filesep,'bin',filesep,lower(computer),filesep,'libcublas.so.',sprintf('%1.1f',g.ToolkitVersion)];    
            defaultcublas=[];
        elseif ispc
            cublaspath=[];
            defaultcublas=' -lcublas ';
        end
        eval(['mex -silent -largeArrayDims',' -f ''', mxoptsfile,'''',' -I''',pwd,'''',' -I''',project_includepath,'''',' -I''',model_includepath,'''',' -I''',libpath,'''',' LINKLIBS=''$LINKLIBS ',cublaspath,'''',defaultcublas,fullstr,' ''', srcfile,'''']);

    end
     
else
    outcompile = -1;
end
