/*
 * m_thetapar.cu
 *
 * Model-specific constants to be provided by a model developer.
 * Reminder: C is row-major, but Matlab uses column-major!!
 *
 * Defaults are coded below. They are overriden by providing
 * -DXXX=VAL as options to the mex compiler
 */

#ifndef Y_N
#define	Y_N 3
                        // N is the number of outcomes per observation
#endif

#ifndef X_K
#define	X_K 3
                        // K is the number of covariates per observation
#endif

#ifndef THREADS_PER_BLOCK
#define	THREADS_PER_BLOCK 96
                        // the number of threads in each block...
                        // set to integer multiple of 32
#endif

//
// Maps for A, B, H and X, Y
// Note: __constant__  is a CUDA extension to store data on GPU "Constant Memory"
// 
__constant__ int X_K_rt = X_K;
__constant__ int Y_N_rt = Y_N;

__constant__ int Asize_rt = 1;
__constant__ int list2A_rt[2*Y_N*Y_N];
__constant__ int list_Amap_rt[Y_N*Y_N];

__constant__ int Bsize_rt = 1;
__constant__ int list2B_rt[2*X_K*X_K];
__constant__ int list_Bmap_rt[X_K*X_K];

__constant__ int Hsize_rt = 1 ;
__constant__ int list2H_rt[2*Y_N*Y_N];
__constant__ int list_Hmap_rt[Y_N*Y_N];

__device__ inline double m_thetapar_general(  const char       M
                                     , const int        row_pos
                                     , const int        col_pos
                                     , const double *   local_theta
                                     , const int        batchsize
                                     )
{
    double temp = 0;
    switch(M)
    {
        case 'A' :
                break;
        case 'B' :
                break;
        case 'H' :
                break;
        default :
                temp = 0;
    }
    return temp;
}

//
// m_thetapar_H
//
// This function automatically applies H(i,i)=exp(H(i,i))
// ie exponential function is applied to the diagonal elements of H
//

__device__ inline double m_thetapar_H( const int         iter_pos
                                     , const int         row_pos
                                     , const int         col_pos
                                     , const double *    local_theta
                                     , const double *    local_theta_extended
                                     , const int         batchsize
                                     )
{   
    double temp = 0;
    if ( list_Hmap_rt[iter_pos] < 0 )
    {
        temp = *( local_theta_extended + batchsize*(-list_Hmap_rt[iter_pos] - 1) );
    }
    else
    {
        if ( list_Hmap_rt[iter_pos] > 0 )
        {
            if (row_pos == col_pos)   // make entries exponential on the diagonal of H to enforce non-negativity
            {
                // temp = fmax( 1e-16 , *( local_theta + batchsize*(list_Hmap_rt[iter_pos] - 1) ) );
                temp = exp( *( local_theta + batchsize*(list_Hmap_rt[iter_pos] - 1) ) );
            }
            else
            {
                temp = *( local_theta + batchsize*(list_Hmap_rt[iter_pos] - 1) );
            }
        }
        else
        {
            // Special handling goes below - keep m_thetapar_H as small as possible!
            temp = m_thetapar_general(   'H'
                                        ,  row_pos
                                        ,  col_pos
                                        ,  local_theta
                                        ,  batchsize
                                        );
        }
    }
    return temp;
}

__device__ inline double m_thetapar_B( const int         iter_pos
                                     , const int         row_pos
                                     , const int         col_pos
                                     , const double *    local_theta
                                     , const double *    local_theta_extended
                                     , const int         batchsize
                                     )
{   
    double temp = 0;
    if ( list_Bmap_rt[iter_pos] < 0 )
    {
        temp = *( local_theta_extended + batchsize*(-list_Bmap_rt[iter_pos] - 1) );
    }
    else
    {
        if ( list_Bmap_rt[iter_pos] > 0 )
        {
            temp = *( local_theta + batchsize*(list_Bmap_rt[iter_pos] - 1) );
        }
        else
        {
            // Special handling goes below - keep m_thetapar_B as small as possible!
            temp = m_thetapar_general(   'B'
                                        ,  row_pos
                                        ,  col_pos
                                        ,  local_theta
                                        ,  batchsize
                                        );
        }
    }
    return temp;
}

__device__ inline double m_thetapar_A( const int         iter_pos
                                     , const int         row_pos
                                     , const int         col_pos
                                     , const double *    local_theta
                                     , const double *    local_theta_extended
                                     , const int         batchsize
                                     )
{   
    double temp = 0;
    if ( list_Amap_rt[iter_pos] < 0 )
    {
        temp = *( local_theta_extended + batchsize*(-list_Amap_rt[iter_pos] - 1) );
    }
    else
    {
        if ( list_Amap_rt[iter_pos] > 0 )
        {
            temp = *( local_theta + batchsize*(list_Amap_rt[iter_pos] - 1) );
        }
        else
        {
            // Special handling goes below - keep m_thetapar_A as small as possible!
            temp = m_thetapar_general(   'A'
                                        ,  row_pos
                                        ,  col_pos
                                        ,  local_theta
                                        ,  batchsize
                                        );
        }
    }
    return temp;
}