function Ysim = m_mvnsim(X, tsim, varargin)
%  Simulate the multivariate normal model Y = Y*A' + X*B' + E.
%  In this code all predetermined variables are also strictly exogenous.
%
%  Input configuration 1:
%  X        Matrix of exogenous variables  (T x k)
%  tsim     Rows of X used in simulation
%  A        Matrix A  (n x n)
%  B        Matrix B  (n x k)
%  H        Matrix H: H'*H = precision = inv(var(e_t)).
%
%  Input configuration 2:
%  X        Matrix of exogenous variables  (T x k)
%  tsim     Rows of X used in simulation
%  theta    Particle matrix
%  Amap     These three aguments map theta to the A matrix 
%  Alinc    (See u_collectpart)
%  Aname
%  Bmap     These three aguments map theta to the B matrix 
%  Blinc    (See u_collectpart)
%  Bname
%  Hmap     These three aguments map theta to the H matrix 
%  Hlinc    (See u_collectpart)
%  Hname
%
%  Output:
%  Ysim     Simulated Y  (length(tsim) x n)

T = length(tsim);

if length(varargin) == 3
    A = varargin{1};
    B = varargin{2};
    H = varargin{3};
    n = size(B, 1);
    H1 = inv(H);
    A1 = eye(n)-A';
    if rank(A1) < n
        error('eye(%d) - A is rank deficient', n)
    end
    Ysim = (X(tsim,:)*B' + randn(T, n)*H1')/A1;
    
elseif length(varargin) == 10
    theta = varargin{1};
    Amap = varargin{2};
    Alinc = varargin{3};
    Aname = varargin{4};
    Bmap = varargin{5};
    Blinc = varargin{6};
    Bname = varargin{7};
    Hmap = varargin{8};
    Hlinc = varargin{9};
    Hname = varargin{10};
    A = u_3dpart(theta, Amap, Alinc, Aname);
    B = u_3dpart(theta, Bmap, Blinc, Bname);
    H = u_3dpart(theta, Hmap, Hlinc, Hname);
    JN= size(A, 1);
    n = size(Amap, 1);
    Ysim = zeros(T, n, JN);
    for j = 1:JN
        H1 = squeeze(H(:,:,j));
        H1(1:n+1:n^2) = exp(H1(1:n+1:n^2));
        H1 = inv(H1);
        A1 = eye(n)-squeeze(A(:,:,j)');
        if rank(A1) < n
            error('eye(%d) - A is rank deficient, row %d of theta', n, j)
        end
        Ysim(:,:,j) = (X(tsim,:)*squeeze(B(:,:,j))' + randn(T, n)*H1')/A1;
    end
    
else
    error('Function has %d input arguments ~= permitted 5 or 12', ...
        2+length(varargin))

end