function m_monitor(stage)
%  Execute monitoring functions specific to the egarch model but not to
%  the project.
%
%  Intput:
%  stage        Current stage of the SABL algorithm

global C Cpar M

if strcmp(stage, 'startrun') && ~C.simget 
    C.Cphase_method = 'data_whole';
end

p_monitor(stage);

if strcmp(stage, 'startrun') && ~C.simget
    if ~strcmp(C.Cphase_method, 'data_whole')
        error('Model egarch requires C.Cphase_method = data_whole')
    end
    M = u_tidyfields(M);
    %  Data
    [~, k] = u_datacheck;
    if ~isfield(C, 'tfirst')
        C.tfirst = 1;
    end
    if ~isfield(C, 'tlast')
        C.tlast = size(M.data, 1);
    end
    u_pointercheck(M, 'M', k, 'y_pointer', 1);
    u_isfield(M, 'I', true, 'M');
    u_isfield(M, 'K', true, 'M');
    u_is('positiveinteger', 'M.I', M.I);
    u_is('positiveinteger', 'M.K', M.I);
    
    %  Model
    fprintf('\n SABL egarch model\n')
    C.parameters = 2 + 3*(M.K + M.I);
    C.nstates = M.K + 1;
    M.vindex = (1:M.K)';
    M.epsindex = M.K + 1;
    allpars = {'muY', 'sigmaY', 'alpha', 'beta', ...
        'gamma', 'p', 'mu', 'sigma'};
    fieldcheck = true(8,1);
    for j = 1:8
        field = char(allpars{j});
        fieldcheck(j) = u_isfield(M, strcat(field, 'map'));
%         fieldcheck(j) = u_isfield(M.(field), 'map');     
    end
    if ~(all(fieldcheck) || all(~fieldcheck))
        error('Either all M.*map must be specified, or none (default)')
    end
    defaultmap = ~all(fieldcheck);
    if defaultmap
        I = M.I;
        K = M.K;
        M.muYmap = 1; % Linear transformation
        M.sigmaYmap = 2;
        M.alphamap = (3:2+K)';
        M.betamap = M.alphamap + K;
        M.gammamap = M.betamap + K;
        M.pmap = (M.gammamap(K)+1:M.gammamap(K)+I)';
        M.mumap = M.pmap + I;
        M.sigmamap = M.mumap + I;
        C.parameters = M.sigmamap(I);
        M.sigmaYmap = M.sigmaYmap + C.parameters;        %  Nonlinear
        M.alphamap = M.alphamap + C.parameters;      %      :
        M.betamap = M.betamap + C.parameters;        %      :
        M.mumap = M.mumap + C.parameters;            %      :
        M.pmap = M.pmap + C.parameters;              %      :
        M.sigmamap = M.sigmamap + C.parameters;      %  Nonlinear
        M.muYlinc = [.001; zeros(C.parameters-1, 1)]; %  Linear
    end
    parlength = [1; 1; M.K; M.K; M.K; M.I; M.I; M.I];
    for j = 1:8
        field = char(allpars{j});
        fieldmap = strcat(field, 'map');
        fieldlinc = strcat(field, 'linc');
        M = u_blankfields(M, fieldlinc);
        M = u_mapcheck(M, 'M', fieldmap, [parlength(j), 1], C.parameters, ...
            fieldlinc);
    end
    
    %  Prior
    if u_isfield(M, 'prior')
    else   
        if ~defaultmap
            error('Default prior requires default parameter map')
        end
        for j = 1:8
            field = char(allpars{j});  % e.g. 'muY'
            fieldprior = strcat(field, 'prior');  % e.g., 'muYprior'
            priorstruc = strcat('M.', fieldprior);  % e.g., 'M.muYprior'
            u_isfield(M, fieldprior, true, 'M');
            priorstrucmean = strcat(priorstruc, 'mean');
            priorstrucstd = strcat(priorstruc, 'std');
            u_isfield(M.(fieldprior), 'mean', true, priorstruc);
            u_isfield(M.(fieldprior), 'std', true, priorstruc);
            u_is('realfinite', priorstrucmean, M.(fieldprior).mean);
            u_is('positivereal', priorstrucstd, M.(fieldprior).std);
        end
        M.prior{1}.mean = [M.muYprior.mean; M.sigmaYprior.mean; ...
            repmat(M.alphaprior.mean, M.K, 1); ...
            repmat(M.betaprior.mean, M.K, 1); ...
            repmat(M.gammaprior.mean, M.K, 1);
            repmat(M.muprior.mean, M.I, 1);
            repmat(M.pprior.mean, M.I, 1);
            repmat(M.sigmaprior.mean, M.I, 1)];
        M.prior{1}.std = [M.muYprior.std; M.sigmaYprior.std; ...
            repmat(M.alphaprior.std, M.K, 1); ...
            repmat(M.betaprior.std, M.K, 1); ...
            repmat(M.gammaprior.std, M.K, 1);
            repmat(M.muprior.std, M.I, 1);
            repmat(M.pprior.std, M.I, 1);
            repmat(M.sigmaprior.std, M.I, 1)];
        M.prior{1}.columns = (1:C.parameters)';
        M.prior{1} = u_prior_linearsetup(M.prior{1});
    end
    
elseif strcmp(stage, 'initialize')
    C.intensity = u_allocate([C.tlast,1]);
    Cpar.states_in = u_allocate([C.JNwork, C.nstates]);
    Cpar.states_out = u_allocate([C.JNwork, C.nstates]);

elseif strcmp(stage, 'startCphase')
    Cpar.states_in = Cpar.states_out;
    
elseif strcmp(stage, 'whileCphase')
    if strcmp(C.Cphase_method, 'data_whole')
        Cpar.states_in = Cpar.states_out;
    else  %  Revise with future implementation
        error('Cphase_method %s not handled in m_monitor',C.Cphase_method)
    end
    
elseif strcmp(stage, 'startMphase') || strcmp(stage, 'whileMphase')
    if strcmp(C.Mphase_method, 'MGRW_simple') || ...
            strcmp(C.Mphase_method, 'MGRW_blocked')
        Cpar.states_in = u_allocate([C.JNwork, C.nstates]);
    else  %  Revise with future implementation
        error('Mphase_method %s not handled in m_monitor',C.Mphase_method)
    end
        
end
