function [logpy, v, eps] = m_egarch1(muY, sigY, alpha, beta, gamma, ...
    p, mu, sig, y, intensity, vlag_in, epslag_in)
%  Evaluate the log p.d.f. of the EGARCH model of Durham
%  and Geweke (2013) for a single value of the parameters.
%
%  Inputs:
%  muY        Mean of y_t  (scalar)
%  sigY       Standard deviation of y_t  (scalar)
%  alpha      Alpha parameters of volatility dynamics (K x 1)
%  beta       Beta parmaeters of volatility dynamics (K x 1)
%  gamma      Gamma parameters of volatility dynamics (K x 1)
%  p          Parameters p_i of epsilon normal mixture (I x 1)
%  mu         Parameters mu_i of epsilon normal mixture (I x 1)
%  sig        Parameters sig_i of epsilon normal mxiture (I x 1)
%  y          Observed time series y_1:t2  (t2 x 1)
%  intensity  Intensity vector
%  vlag_in    v_t1-1.  By convention vlag = 0 if t1 = 1.  (K x 1)
%  epslag_in  epsilon_t1-1.  By convention epslag = 0 if t1 = 1.  (Scalar)
%
%  Outputs:
%  logpy      log p(y_t1:t2 | y_1:t1-1, parameters)  (Scalar)
%  v          v_t2  (K x 1)
%  eps        epsilon_t2  (Scalar)

shift = sqrt(2/pi);

logpy = 0;
tt = find(intensity > 0);
vlag = vlag_in;
epslag = epslag_in;
for i = 1:length(tt);
    t = tt(i);
    v = alpha.*vlag + beta*(abs(epslag) - shift) + gamma*epslag;
    h = sigY*exp(0.5*sum(v));
    eps = (y(t) - muY)/h;
    logpy = logpy + intensity(t)*log(dot(p,normpdf(eps,mu,sig))/h);
    vlag = v;
    epslag = eps;
end

end   