function p_monitor(stage)
%  Set up the information structure for this project.
%  Default fields in the M structure are for illustration only, so
%  they will be re-set, in general.
%  Set up any additional infrastructure specific to the project,
%  creating a P structure to do so if desired.
%
%  The following code generates a simpler data set of Klein model 1;
%
%  In this project, the precision Matrix is assumed to follow Wishart
%  distribution.
%
%  M.B_mean      The prior mean of the non-zero elements in B
%  M.B_std       The prior std of the non-zero elements in B
%  M.S_prec      The scale matrix of Wishart distribution
%  M.prior.S_df  The degrees of freedom of Wishart distribution
%
%  Note that the elements of A is the combination of elements of matrix B
%  in this project
%
%  The default data generating process is a simpler verson of
%  Klein model 1.
%  Y = A*Y + B*X + E
%  A = (0 0 b1-b2; 0 0 b3; b4 0 0)
%  B = (b1 b2 0; 0 b3 0; 0 0 b4)

global C Cpar E M P
persistent step

if P.test  %  --------------------------------------------------------
    if strcmp(stage, 'open')
        E.gpu = (P.gpu == 2);
        E.pus = P.workers;
        
        step = P.step;
        if step == 2
            E.simulation_get = 'simfile';
        end
        
    elseif strcmp(stage, 'startrun')
        
        C.Cphase_method = char(P.Cphase_method{P.Cphase});
        C.Mphase_method = char(P.Mphase_method(P.Mphase));
        C.twopass = (P.twopass == 2);
        C.detail_level = 7;
        C.J = 2^4;
        C.N = 512;
        
        % Below is project specifi; User or Modeller have to privode values for
        % the following fields.
        M.specification = 'FI';
        load Mdata
        M.data = u_setdatatype(Mdata);
        if step <= 1
            M.data = M.data(1:100, :);
        else
            M.data = M.data(1:200, :);
        end
        M.T = size(M.data, 1);
        
        M.sufficient_statistics = false;
        
        M.y_pointer = [4 5 6];
        M.x_pointer = [1 2 3];
        
        % Specify the number of variables needed to be estimated
        M.A_counter = 0;
        M.B_counter = 4;
        M.H_counter = 6;
        
        M.Amap = [0 0 -1; ...
            0 0 3; ...
            4 0 0];
        
        M.Alinc = [1 -1 0 0 0 0 0 0 0 0]';
        
        M.Bmap = [1 2 0; ...
            0 3 0; ...
            0 0 4];
        
        M.Blinc = [];
        
        M.Hmap = [5 6 8; ...
            0 7 9; ...
            0 0 10];
        
        C.tlast = size(M.data, 1);
        C.parameters = 10;
        
        M.specification = 'FI';
        M.meanB = u_allocate([1, 4], 'default', 'zeros');
        M.stdB  = u_allocate([1, 4], 'default', 'ones');
        
        %  Part 4: Set up the prior
        M.prior{1}.distribution = 'normal';
        M.prior{1}.mean = M.meanB';
        M.prior{1}.std = M.stdB';
        M.prior{1}.columns = (1:4)';
        M.prior{1} = u_prior_linearsetup(M.prior{1});
        
        M.prior{2}.variance = eye(3);
        M.prior{2}.df = 5;
        M.prior{2}.columns = (5:10)';
        M.prior{2} = u_prior_wishartsetup(M.prior{2});
        
    elseif strcmp(stage, 'finish')
   
        P.tmean = u_mean(Cpar.theta, 1, E.pus>1);
        P.tstd  = u_std(Cpar.theta, 1, E.pus>1);
        P.tnse  = u_nse(Cpar.theta, C.N, E.pus>1);
        P.trne  = u_rne(Cpar.theta, C.N, E.pus>1);
        if ~strcmp(C.Cphase_method, 'anneal_optimize')
            P.logml1 = C.logml1;
            P.logmlnse = C.logmlnse;
        else
            P.logml1 = 0;
        end
        
    elseif strcmp(stage, 'close')
        if step == 1
            E.simulation_record = 'simfile';
        end
    end
    
else  %  -----------------------------------------------------------------
    
end

end



