function p_monitor(stage)
%  Set up the information structure for this project. Default fields in the
%  M structure are for illustration only, so they will be re-set, in
%  general. Set up any additional infrastructure specific to the project,
%  creating a P structure to do so if desired.
%
%  Intput:
%  stage        Current stage of the SABL algorithm

%  The following code loads the data set for Klein model 1. The data is
%  collected from Theil (1971). The detailed data description can be found
%  therein.
%
%  In this project, the precision Matrix is assumed to follow Wishart
%  distribution.     
%
%  Note that the elements of A is the combination of elements of matrix B
%  in this project
global C Cpar E M

if strcmp(stage, 'open')
    % E.gpu = true;
    % E.pus = 2;
    
elseif strcmp(stage, 'startrun')
    
    % Change the number of columns of B
    no_row_A = 6;
    no_col_B = 4;
    
    M.y_pointer = (1:no_row_A) + no_col_B;
    M.x_pointer = 1:no_col_B;
    M.y_pointer = M.y_pointer';
    M.x_pointer = M.x_pointer';
    
    M.specification = 'FI';
    A = [0 0 0.5; 0 0 1; 1 0 0];
    A = blkdiag(A, zeros(no_row_A-3, no_row_A-3));
    M.AA = A;
    
    B = [1 0.5 0 zeros(1, no_col_B - 3); ...
         0   1 0 zeros(1, no_col_B - 3); ...
         0   0 1 zeros(1, no_col_B - 3); ...
         zeros(no_row_A - 3, no_col_B)];
    M.BB = B;
    M.data(:, M.x_pointer) = randn(1000, no_col_B);
    
    for i = 1:size(M.data, 1);
        M.data(i, M.y_pointer) = M.data(i, M.x_pointer)*(B')/((eye(no_row_A)-A)') ...
            + randn(1, no_row_A)/((eye(no_row_A)-A)');
    end
    M.data = u_setdatatype(M.data);
    
    % Specify the number of variables needed to be estimated
    
    % Specify the prior mean of the elements of B(theta)   
    M.Amap = [0 0 -1; ...
              0 0  3; ...
              4 0  0];
    Ablock = zeros(no_row_A -3, no_row_A -3);
    Acounter = 0;
    for i = 1:(no_row_A -3)
        for j = 1:(no_row_A -3)
            if (j+1) == i
                Acounter = Acounter + 1;
                Ablock(j, i) = 4 + Acounter; 
            end
        end
    end
    M.Amap = blkdiag(M.Amap, Ablock);
    disp(M.Amap)
    
    M.Bmap = [1 2 0 4+Acounter+(1:(no_col_B - 3)); ...
              0 3 0 4+Acounter+(no_col_B - 3)+(1:(no_col_B - 3)); ...
              0 0 4 4+Acounter+2*(no_col_B - 3)+(1:(no_col_B - 3))];
    
    for i = 1:(no_row_A-3);
        M.Bmap = [M.Bmap; ...
            4+Acounter+3*(no_col_B - 3)+(i-1)*no_col_B+(1:no_col_B)];
    end
    
    no_AB = Acounter + nnz(M.Bmap);
    Hblock = triu(ones(no_row_A, no_row_A),0);
    Hcounter = 0;
    for i = 1:no_row_A
        for j = 1:no_row_A
            if Hblock(j, i) == 1
                Hcounter = Hcounter + 1;
                Hblock(j, i) = no_AB + Hcounter; 
            end
        end
    end
    M.Hmap = Hblock;
    
    C.tfirst = 1;
    C.tlast = size(M.data, 1);
    C.parameters = no_AB + nnz(M.Hmap);
    
    M.Alinc = [1 -1 zeros(1, C.parameters - 2)]';  
    M.Blinc = []; 
    M.Hlinc = [];

    M.meanAB = zeros(1, no_AB);
    M.stdAB  = ones(1, no_AB);
    %  Part 4: Set up the prior
    M.prior{1}.distribution = 'normal';
    M.prior{1}.mean = M.meanAB';
    M.prior{1}.std = M.stdAB';
    M.prior{1}.columns = (1:no_AB)';
    M.prior{1} = u_prior_linearsetup(M.prior{1});
    
    M.prior{2}.distribution = 'Wishart';
    M.prior{2}.variance = eye(no_row_A);
    M.prior{2}.df = no_row_A+3;
    M.prior{2}.columns = no_AB+(1:nnz(M.Hmap))';
    M.prior{2} = u_prior_wishartsetup(M.prior{2});
    
elseif strcmp(stage, 'finish')
    
    A = M.AA;
    B = M.BB;
    H = eye(length(M.y_pointer));
    
    [Aest, M.Amap] = u_collectpart(Cpar.theta, M.Amap, M.Alinc, 'A');
    M.Amap = u_setdatatype(M.Amap, 'cpu');
    
    Amean = u_allocate([M.n, M.n]);
    Astd  = u_allocate([M.n, M.n]);
    for i = 1:M.n
        for j = 1:M.n
            if M.Amap(i,j) == 0
                Amean(i,j) = 0;
                Astd(i,j) = 0;
            else
                Amean(i,j) = u_mean(Aest(:,M.Amap(i,j)), 1, E.pus>1);
                Astd(i,j) = u_std(Aest(:,M.Amap(i,j)), 1, E.pus>1);
            end
        end
    end
    disp('A posterior mean');  disp(Amean)
    disp('A posteror std');  disp(Astd)
    disp('t stats');  disp((Amean-A)./Astd)

    Bmean = u_allocate([M.n, M.k]);
    Bstd  = u_allocate([M.n, M.k]);
    for i = 1:M.n
        for j = 1:M.k
            if M.Bmap(i,j) == 0
                Bmean(i,j) = 0;
                Bstd(i,j) = 0;
            else
                Bmean(i,j) = u_mean(Cpar.theta(:,M.Bmap(i,j)), 1, E.pus>1);
                Bstd(i,j) = u_std(Cpar.theta(:,M.Bmap(i,j)), 1, E.pus>1);
            end
        end
    end
    disp('B posterior mean');  disp(Bmean)
    disp('B posteror std');  disp(Bstd)
    disp('t stats');  disp((Bmean-B)./Bstd)
    
    Hmean = u_allocate([M.n, M.n]);
    Hstd  = u_allocate([M.n, M.n]);
    for i = 1:M.n
        for j = i:M.n
            xx = Cpar.theta(:, M.Hmap(i,j));
            if i == j
                xx = exp(xx);
            end
            Hmean(i,j) = u_mean(xx, 1, E.pus>1);
            Hstd(i,j) = u_std(xx, 1, E.pus>1);
        end
    end
    disp('H posterior mean');  disp(Hmean)
    disp('H posteror std');  disp(Hstd)
    disp('t stats');  disp((Hmean - H)./Hstd)
end
    
end














